"""Asian Transport Observatory (ATO) provider."""

import logging
from collections import defaultdict
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from itertools import chain, product
from typing import TYPE_CHECKING
from urllib.parse import quote, urlparse

import pandas as pd
import sdmx.model.v21 as m
from sdmx.model import common, v21

from transport_data.util.pluggy import hookimpl
from transport_data.util.pooch import Pooch
from transport_data.util.sdmx import anno_generated

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)

BASE_URL = "https://asiantransportobservatory.org/exportdl?orig=1"

#: List of all "ECONOMY" codes appearing in processed data.
CL_ECONOMY = m.Codelist(
    id="ECONOMY",
    name="Asian Transport Outlook subject economy",
    description="Equivalent to GEO or REF_AREA in other concept schemes.",
)

#: List of all measures (indicators) appearing in processed data.
#:
#: .. todo:: Validate against the master list of indicators; or read from that file and
#:    validate IDs appearing in data files.
CS_MEASURE = m.ConceptScheme(
    id="MEASURE",
    name="Asian Transport Outlook measures (indicators)",
    description="Item are automatically generated by the TDCI tools. Currently no "
    "correspondence with the list provided directly by ATO is checked or enforced.",
)

#: Mapping from short codes for ATO data categories to file names.
FILES = {
    # "ATO National Database Masterlist of Indicators",
    "ACC": (
        "ATO Workbook (ACCESS & CONNECTIVITY (ACC)).xlsx",
        "sha256:21c3b7e662d932f5cc61c22489acb3cf0e8a70200abc2372c7fe212602903fd7",
    ),
    "APH": (
        "ATO Workbook (AIR POLLUTION & HEALTH (APH)).xlsx",
        "sha256:dcec4676c74566712e2771aad0afe196d1db9a3f7630eac1c3dba29d0b7c09f4",
    ),
    "CLC": (
        "ATO Workbook (CLIMATE CHANGE (CLC)).xlsx",
        "sha256:2d582ade3dfe452fb2eedb3cbe9d06ac7167f0bb867e41016c8a1e7aa2efca15",
    ),
    "INF": (
        "ATO Workbook (INFRASTRUCTURE (INF)).xlsx",
        "sha256:8ee1274cd2268c44b1f33d3917b72fc6626897aca28ecbee56c0ba8aa646e89a",
    ),
    "MIS": (
        "ATO Workbook (MISCELLANEOUS (MIS)).xlsx",
        "sha256:c601e9e217e137a6071758f73cac050ea7dae4ff746e4a99d8c3297269175c03",
    ),
    "POL": (
        "ATO Workbook (TRANSPORT POLICY (POL)).xlsx",
        "sha256:fbf23b012590b631239654d255d23ccb70fa717b466be8343a5b0f1e8b4ce720",
    ),
    "RSA": (
        "ATO Workbook (ROAD SAFETY (RSA)).xlsx",
        "sha256:51a6658fa12fcb3ac77298f5908ab343492385ecb1b24602ba21b91dbbcedca5",
    ),
    "SEC": (
        "ATO Workbook (SOCIO-ECONOMIC (SEC)).xlsx",
        "sha256:bc5e4a0006173a53f5b5f283c3b0174566b81842e368a432f25ba563ffcda93b",
    ),
    "TAS": (
        "ATO Workbook (TRANSPORT ACTIVITY & SERVICES (TAS)).xlsx",
        "sha256:3e468c325ab508476d5d06e81d7d0e2c21655b4f3801abf20776812928126bb6",
    ),
}

VERSION = "0.1.0"


@dataclass
class Config:
    """Common configuration for :func:`fetch`, :func:`convert` and others."""

    #: If :any:`True`, fetch from the Zenodo mirror at :data:`POOCH_ZENODO`. Otherwise,
    #: fetch from :data:`POOCH`.
    from_zenodo: bool = True

    #: If :any:`True`, do not fetch anything; only confirm that the files to be fetched
    #: are available.
    dry_run: bool = False


def expand(fname: str) -> str:
    """Callback to expand `fname` into a complete file name for :data:`POOCH`."""
    return FILES.get(fname, (fname, None))[0]


def expand_zenodo(fname: str) -> str:
    """Callback to expand `fname` into a complete file name for :data:`POOCH_ZENODO`."""
    POOCH_ZENODO._ensure_doi_registry()

    # Identify all files with names containing "(fname)"
    matches = set(k for k in POOCH_ZENODO.registry if f"({fname})" in k)

    try:
        # Most recent file sorted by date
        return sorted(matches)[-1]
    except IndexError:  # pragma: no cover
        raise KeyError(fname)


def _make_url(fname: str) -> str:
    return f"{BASE_URL}&filename={quote(expand(fname))}"


#: Pooch to fetch data from the ATO website.
POOCH = Pooch(
    module=__name__,
    base_url=BASE_URL,
    registry={v[0]: v[1] for v in FILES.values()},
    urls={expand(key): _make_url(key) for key in FILES.keys()},
    expand=expand,
)

#: Pooch to fetch data from the Zenodo mirror at
#: https://doi.org/10.5281/zenodo.14913729.
POOCH_ZENODO = Pooch(
    module=f"{__name__}-zenodo",
    base_url="doi:10.5281/zenodo.15232577/",
    expand=expand_zenodo,
)


def convert(part: str, config: Config | None = None, **kwargs) -> None:
    """Convert `part` of the ATO National Database to SDMX-ML and store."""
    from tqdm import tqdm

    from transport_data import STORE

    # Ensure the source file is available locally
    paths = fetch(part, config=config, **kwargs)
    assert 1 == len(paths), paths
    ef = pd.ExcelFile(paths[0], engine="openpyxl")

    # Maximum length of a sheet name, for formatting the description
    L = max(map(lambda v: len(str(v)), ef.sheet_names))

    for sheet_name in (progress := tqdm(ef.sheet_names)):
        if sheet_name == "TOC":
            # Skip the TOC
            continue
        progress.set_description(f"Process sheet {sheet_name!r:>{L}}")

        df, annos = read_sheet(ef, str(sheet_name))
        ds = convert_sheet(df, annos)

        # Write the DSD and DFD
        STORE.set(ds.described_by)
        STORE.set(ds.structured_by)
        # Write the data itself, to SDMX-ML and CSV
        STORE.set(ds)

    # Write the lists of "Economy" codes and measures/concepts accumulated while
    # converting
    a = get_agencies()[0]
    for obj in (CL_ECONOMY, CS_MEASURE):
        obj.maintainer = a
        obj.version = VERSION
        STORE.set(obj)


def convert_sheet(df: pd.DataFrame, aa: m.AnnotableArtefact):
    """Convert `df` and `aa` from :func:`read_sheet` into SDMX data structures."""
    # Prepare an empty data set, associated structures, and a helper function
    ds, _make_obs = prepare(aa)

    # - Validate values in "Economy Code", "Economy Name" against CL_ECONOMY; then keep
    #   only the former as "ECONOMY".
    # - Melt into long format (one observation per row), preserving remarks columns.
    # - Drop NA values in the "OBS_VALUE" column
    data = (
        df.pipe(validate_economy)
        .melt(
            id_vars=["ECONOMY"] + (ds.eval_annotation("remark-cols") or []),
            var_name="TIME_PERIOD",
            value_name="OBS_VALUE",
        )
        .dropna(subset=["OBS_VALUE"])
    )

    # Convert rows of `data` into SDMX Observation objects
    ds.obs.extend(_make_obs(row) for _, row in data.iterrows())
    assert len(ds) == len(data)

    ds.pop_annotation(id="remark-cols")

    return ds


ATTRIBUTES: set[str] = set()


def _is_url(value: str) -> bool:
    """Return :any:`True` if `value` contains a URL."""
    return urlparse(value).path != value


def dataset_to_metadata_reports(
    ds: "v21.DataSet", msd: "v21.MetadataStructureDefinition"
) -> Iterable["v21.MetadataReport"]:
    """Convert the attributes of ATO `ds` to 1 or more :class:`.MetadataReport`.

    The metadata reports conform to the TDC metadata structure (
    :func:`.org.metadata.get_msd`).

    If `ds` contains per-series values for attributes named “Source”, “Source
    (2024-11)”, or similar, then additional metadata reports are generated, one for each
    series (=GEO, or ‘economy’) and each distinct upstream source indicated by these
    attribute values.
    """
    import sdmx.urn
    from pycountry import countries

    from transport_data import STORE
    from transport_data.org.metadata import make_ra, make_tok

    # Retrieve the DSD and DFD
    assert ds.structured_by and ds.structured_by.urn
    urn_dsd = sdmx.urn.normalize(ds.structured_by.urn)
    dsd = STORE.get(urn_dsd)
    dfd = STORE.get(urn_dsd.replace("DataStructure", "Dataflow"))
    # Assign the loaded DSD to the DFD's `structure` attribute, as these are loaded
    # separately
    dfd.structure = dsd

    # - Create a TargetObjectKey to associate the metadata report with the data flow
    #   definition
    # - Create the report itself
    mdr = v21.MetadataReport(attaches_to=make_tok(dfd))

    # Add the "DATAFLOW" metadata attribute, containing the same information
    mdr.metadata.append(make_ra("DATAFLOW", f"ATO:{dfd.id}"))

    # Map:
    # - from ATO IDs for attributes attached to data sets
    # - to TDC IDs for metadata attributes
    for attribute_id, mda_id in (
        ("SOURCE", "DATA_PROVIDER"),
        ("UNITS", "UNIT_MEASURE"),
        ("WEBSITE", "URL"),
    ):
        # Retrieve the (data) attribute value
        value = str(ds.attrib[attribute_id].value)
        # Convert to a reported (metadata) attribute; append to the report
        value = format_data_provider(value) if mda_id == "DATA_PROVIDER" else value
        mdr.metadata.append(make_ra(mda_id, value))

    # Retrieve the MEASURE from CS_MEASURE
    cs_measure = STORE.get("ConceptScheme=ATO:MEASURE(0.1.0)")
    measure_concept = cs_measure[dsd.id]
    # Construct a metadata attribute
    mdr.metadata.append(make_ra("MEASURE", measure_concept.name))

    # Assemble description
    description_lines = [
        f"Original (ATO) description of {dsd.id!r}:",
        str(measure_concept.description),
    ]

    ra_comment = make_ra("COMMENT", "Metadata generated by transport_data.tools.")
    mdr.metadata.append(ra_comment)

    geo_all = set()  # All GEO/REF_AREA for which the data set contains data
    sources = defaultdict(set)  # Mapping from GEO → "Source…" attribute values

    for observation in ds.obs:
        # Identify GEO/REF_AREA for which the data set contains data
        # Key value for the 'ECONOMY' dimension of `observation`
        geo_alpha_3 = observation.key["ECONOMY"].value
        # Convert from ISO 3166-1 alpha-3 (used by ATO) to alpha-2 (SDMX convention)
        geo_alpha_2 = countries.lookup(geo_alpha_3).alpha_2
        geo_all.add(geo_alpha_2)

        # Iterate over source attributes
        for av in observation.attached_attribute.values():
            ATTRIBUTES.add(av.value_for.id)  # DEBUG Track all attributes

            if not av.value_for.id.startswith("Source"):
                continue
            sources[geo_alpha_2].add(av.value)

    # Append a sorted list of GEO to the description
    description_lines.extend(
        ["", "The data contain non-null values for GEO:"] + sorted(geo_all)
    )

    # Finalise description
    mdr.metadata.append(make_ra("DATA_DESCR", "\n".join(description_lines)))

    # List of 1 or more metadata reports to return
    result = [mdr]

    # Generate additional metadata reports for upstream data
    for geo, source in chain(*[product([g], sorted(s)) for g, s in sources.items()]):
        dfd_geo = v21.DataflowDefinition(id=f"{geo}_{hash(dfd.id + source)}")
        mdr_geo = v21.MetadataReport(
            attaches_to=make_tok(dfd_geo), metadata=[ra_comment]
        )

        # Store the source attribute value as either the URL or DATA_PROVIDER
        # reported metadata attribute, depending on whether it is a URL
        mda_id = "URL" if _is_url(source) else "DATA_PROVIDER"
        mdr_geo.metadata.append(make_ra(mda_id, source))

        desc = (
            f"Upstream source(s) for 'ATO:{dfd.id}' (“{measure_concept.name}”) "
            f"data for GEO={geo!r}."
        )
        mdr_geo.metadata.append(make_ra("DATA_DESCR", desc))

        result.append(mdr_geo)

    return result


def fetch(*parts: str, config: Config | None = None, **kwargs) -> list["Path"]:
    """Fetch source data files for one or more parts of the ATO National Database.

    Parameters
    ----------
    parts
        If no positional args are given, all of the keys from :data:`FILES` are used.
    """
    if not parts:
        parts = tuple(sorted(FILES))

    config = config or Config(**kwargs)

    pooch = POOCH_ZENODO if config.from_zenodo else POOCH

    result = []

    if config.dry_run:
        for p in parts:
            log.info(f"Valid url for ATO part {p!r}: {pooch.is_available(p)}")
            result.append(pooch.path_for(p))
    else:
        result.extend(pooch.fetch(p) for p in parts)

    return result


def format_data_provider(value: str) -> str:
    """Format the ATO “Source” data attribute as TDC ``DATA_PROVIDER`` metadata.

    This makes more explicit how the ATO has handled upstream data.
    """
    if value.startswith("ATO analysis"):
        return value.replace("ATO analysis", "Asian Transport Outlook (ATO) analysis")
    elif value.startswith("Estimated"):
        return value.replace("Estimated", "Asian Transport Outlook (ATO); estimated")
    elif value.startswith("Calculated"):
        return "Asian Transport Outlook (ATO) calculation"
    else:
        return value + "—republished by ATO"


@hookimpl
def get_agencies():
    a = m.Agency(
        id="ATO",
        name="Asian Transport Observatory",
        description="""See https://asiantransportobservatory.org/aboutus/""",  # noqa: E501
        contact=[
            m.Contact(email=["info@asiantransportobservatory.org"]),
            m.Contact(name="Sudhir Gota", email=["sudhirgota@gmail.com"]),
            # NB Affiliated with ADB, which supports ATO
            # m.Contact(name="James Leather", email=["jleather@adb.org"]),
            # NB No longer formally affiliated with ATO
            # m.Contact(name="Cornie Huizenga", email=["chuizenga@cesg.biz"]),
        ],
    )

    return (a,)


@hookimpl
def provides():
    return (
        "Codelist=TDCI:CL_ATO_ECONOMY",
        "ConceptScheme=TDCI:CS_ATO_MEASURE",
    )


def prepare(aa: m.AnnotableArtefact) -> tuple[m.DataSet, Callable]:
    """Prepare an empty data set and associated structures."""
    # Measure identifier and description
    measure_id = (
        str(aa.pop_annotation(id="INDICATOR-ATO-CODE").text)
        .replace("(", "_")
        .replace(")", "")
    )
    measure_name = aa.pop_annotation(id="INDICATOR").text
    measure_desc = aa.pop_annotation(id="DESCRIPTION").text

    # Measure concept
    c = m.Concept(id=measure_id, name=measure_name, description=measure_desc)
    if c in CS_MEASURE:
        # Avoid duplicating items
        log.info(f"Skip updating {CS_MEASURE[measure_id]!r} with {c!r}")
        c = CS_MEASURE[measure_id]
    else:
        # Extend the "Measure" concept scheme
        CS_MEASURE.append(c)

    # Data structure definition with an ID matching the measure
    # NB here we set ATO as the maintainer. Precisely, ATO establishes the data
    #    structure, but TDCI is maintaining the SDMX representation of it.
    ma_args = dict(maintainer=get_agencies()[0], version=VERSION)
    dsd = m.DataStructureDefinition(id=measure_id, **ma_args)
    anno_generated(dsd)

    dfd = m.DataflowDefinition(id=measure_id, structure=dsd, **ma_args)

    pm = m.PrimaryMeasure(id="OBS_VALUE", concept_identity=c)
    dsd.measures.append(pm)

    # Dimensions
    dsd.dimensions.extend(m.Dimension(id=d) for d in ("ECONOMY", "TIME_PERIOD"))

    # Convert annotations to DataAttributes. "NoSpecifiedRelationship" means that the
    # attribute is attached to an entire data set (not a series, individual obs, etc.).
    da = {}  # Store references for use below
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        da[a.id] = m.DataAttribute(id=a.id, related_to=m.NoSpecifiedRelationship())
        dsd.attributes.append(da[a.id])

    _PMR = m.PrimaryMeasureRelationship  # Shorthand

    # Convert remark column labels to DataAttributes. "PrimaryMeasureRelationship" means
    # that the attribute is attached to individual observations.
    for name in aa.eval_annotation("remark-cols"):
        dsd.attributes.append(m.DataAttribute(id=name, related_to=_PMR()))

    # Empty data set structured by this DSD
    ds = m.DataSet(
        described_by=dfd,
        structured_by=dsd,
        annotations=[aa.get_annotation(id="remark-cols")],
    )
    # FIXME Disabled for https://github.com/transport-data/tools/pull/50; sdmx1 v2.22.0
    #       errors trying to read the generated DataMessage
    # anno_generated(ds)

    # Convert temporary annotation values to attributes attached to the data set
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        ds.attrib[a.id] = m.AttributeValue(value=str(a.text), value_for=da[a.id])

    def _make_obs(row):
        """Helper function for making :class:`sdmx.model.Observation` objects."""
        key = dsd.make_key(m.Key, row[[d.id for d in dsd.dimensions]])

        # Attributes
        attrs = {}
        for a in filter(lambda a: isinstance(a.related_to, _PMR), dsd.attributes):
            # Only store an AttributeValue if there is some text
            value = row[a.id]
            if not pd.isna(value):
                attrs[a.id] = m.AttributeValue(value_for=a, value=str(value))

        return m.Observation(
            dimension=key, attached_attribute=attrs, value_for=pm, value=row[pm.id]
        )

    return ds, _make_obs


def read_sheet(
    ef: pd.ExcelFile, sheet_name: str
) -> tuple[pd.DataFrame, m.AnnotableArtefact]:
    """Read a single sheet.

    This function handles the particular layout of sheets in files like those listed in
    :data:`FILES`. These combine data and metadata.

    - Row 1 is a title row.
    - Cell range A2:B10 contain a set of metadata fields, with the field name in column
      A and the value in column B.
    - Rows 11:13 contain no data or metadata; only a link back to a table of contents
      sheet.
    - Row 14 contains a label "Series" centre-spanned across
    - Row 15 contains column labels, described below.
    - Row 16 and onwards contain data, followed by two blank rows, and two rows with
      attribution/acknowledgements.

    - Columns labeled (i.e. in row 15) "Economy Code" and "Economy Name" contain codes
      and names, respectively, for the geographic units.
    - Columns with numeric labels describe time periods, specifically years, that are
      part of observation keys.
    - Some sheets have additional columns with non-numeric labels like "Remarks",
      "Source (2022-04)", etc.; these give annotations applying to the observations on
      the same row (i.e. for a single "Economy Code" and 1 or more time periods).

    .. note:: Sheets in the ``POL`` category have a different format.
    """
    # Read metadata section
    meta_df = ef.parse(
        sheet_name, header=None, skiprows=1, nrows=9, usecols="A:B"
    ).dropna(how="all")

    # Convert metadata rows to a collection of temporary annotations
    aa = m.AnnotableArtefact()
    for _, data in meta_df.iterrows():
        title = data.iloc[0].rstrip(":")
        anno_id = title.upper().replace(" ", "-")
        aa.annotations.append(
            m.Annotation(
                id=anno_id,
                title=title,
                text=data.iloc[1]
                if isinstance(data.iloc[1], str)
                else repr(data.iloc[1]),
            )
        )

    # Read data section
    df = ef.parse(sheet_name, skiprows=14, skipfooter=2).dropna(how="all")

    # Identify data columns: those with numeric labels
    data_col_mask = list(map(str.isnumeric, df.columns))

    # Handle values not parsed by pd.ExcelFile.parse(). Some cells have values like
    # "12,345.6\t", which are not parsed to float.
    #
    # - Strip trailing whitespace.
    # - Remove thousands separators ("," in 2022-10-17 edition; " " in 2024-05-20
    #   edition) and whitespace before the decimal separator ("10860 .6", 2024-05-20
    #   edition).
    # - RSA-RSI-007_1 contains erroneous values like "55.0g", likely intended to be
    #   "55.0%"; correct this.
    # - Replace miscellaneous erroneous values appearing since the 2024-05-20 edition:
    #   - "-", "n/a", "N/Appl." (both indicating no data)
    #   - "_" (likely erroneous)
    #   - "long ton" (erroneous)
    # - Some sheets contain columns with text labels in miscellaneous/undocumented
    #   formats, not supported by this function:
    #   - SEC-SEG-009: "land", "port"
    #   - MIS-SUM-002: "A", "B", "C", "D"
    #
    #   TODO Ask ATO to provide these data as SDMX, or extend code to convert
    # - Finally, convert to float.
    dtypes = df.loc[:, data_col_mask].dtypes  # Dtypes of data columns only
    for col, _ in filter(lambda x: x[1] != "float", dtypes.items()):
        assert isinstance(col, str)
        df[col] = (
            df[col]
            .str.strip()
            .str.replace(r"(\d)[, ]([\d\.])", r"\1\2", regex=True)
            .str.replace(r"^([\d\.]+)g", r"\1", regex=True)
            .str.replace(r"^(-|long ton|N/Appl\.|n/a|_)$", "NaN", regex=True)
            .str.replace(r"^(land|port|A|B|C|D)$", "NaN", regex=True)
            .astype(float)
        )

    # Identify remarks columns: any entries at the *end* of `df.columns` with non-
    # numeric labels. Use the index of last data column, counting backwards.
    try:
        N = list(reversed(data_col_mask)).index(True)
    except ValueError:
        N = 0

    # Label(s) of any remarks columns
    remark_cols = df.columns.tolist()[-N:] if N else []
    # Store a list of these as another annotation
    aa.annotations.append(m.Annotation(id="remark-cols", text=repr(remark_cols)))

    return df, aa


def validate_economy(df: pd.DataFrame) -> pd.DataFrame:
    """Validate codes for the "ECONOMY" dimension of `df` against :data:`CL_ECONOMY`.

    - Every unique pair of (Economy Code, Economy Name) is converted to a
      :class:`~sdmx.model.common.Code`.
    - These are added to :data:`CL_ECONOMY`. If a Code with the same ID already exists,
      it is checked for an exact match (name, description, etc.)
    - The "Economy Code" column of `df` is renamed "ECONOMY", and contains only values
      from :data:`CL_ECONOMY`. The "Economy Name" column is dropped.
    """
    ec, en = "Economy Code", "Economy Name"

    def _make_code(row: pd.Series):
        return common.Code(id=row[ec], name=row[en], parent=CL_ECONOMY)

    codes = df[[ec, en]].sort_values(ec).drop_duplicates().apply(_make_code, axis=1)
    for c in codes:
        try:
            CL_ECONOMY.append(c)
        except ValueError:
            assert CL_ECONOMY[c.id].compare(c, verbose=True), (
                f"Existing {CL_ECONOMY[c.id]!r} does not match {c!r}"
            )

    return df.rename(columns={ec: "ECONOMY"}).drop(columns=[en])



"""CLI for :mod:`.oica`.

.. runblock:: console

   $ tdc oica --help

.. runblock:: console

   $ tdc oica convert-all --help

.. runblock:: console

   $ tdc oica fetch --help

"""

import click


@click.group("oica", short_help="OICA provider.")
def main():
    """International Organization of Motor Vehicle Manufacturers (OICA) provider."""


@main.command("convert-all")
def convert_all():
    """Convert all data to SDMX."""
    from . import convert

    for measure in "SALES", "STOCK":
        convert(measure)


@main.command("fetch")
@click.option("--go", is_flag=True, help="Actually fetch.")
def fetch_cmd(go):
    """Fetch original data files."""
    from . import fetch

    fetch(dry_run=not go)



"""IPCC structural metadata."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sdmx.model.common


def gen_cl_T311(**kwargs) -> "sdmx.model.Common.Codelist":
    """Generate a code list from the GNGGI, Volume 2, Table 3.1.1.

    The generated code list's URN ends with ``Codelist=TDCI:CL_IPCC_2006_V2_T3.1.1(…)``.

    .. todo:: Expand to include 'Explanation' text from the table as descriptions for
       codes.

    .. todo:: Include internationalized texts (names, descriptions) from the Arabic,
       Chinese, French, Russian, and/or Spanish versions of the documents.
    """
    from sdmx.model.common import Code, Codelist

    cl = Codelist(
        id="CL_IPCC_2006_V2_T3.1.1",
        name="Detailed sector split for the Transport sector",
        description="""Transcribed from 2006 IPCC Guidelines for National Greenhouse Gas Inventories — Volume 2: Energy — Chapter 3: Mobile Combustion — Table 3.1.1, using the file https://www.ipcc-nggip.iges.or.jp/public/2006gl/pdf/2_Volume2/V2_3_Ch3_Mobile_Combustion.pdf, as linked from https://www.ipcc-nggip.iges.or.jp/public/2006gl/vol2.html.

This version includes the 'Explanation' text from the table as the description for individual codes, but at the moment only for the code 1 A 3. For others, see the source document.""",
        **kwargs,
    )

    # The codes have well-formed, hierarchical IDs, so it is possible to infer the ID of
    # the parent code, if it exists.
    def _c(id_, name, description=None):
        """Shorthand for adding to `cl`."""
        try:
            parent = cl[" ".join(id_.split()[:-1])]
        except KeyError:
            parent = None

        cl.append(Code(id=id_, name=name, description=description, parent=parent))

    _c(
        "1 A 3",
        "TRANSPORT",
        """Emissions from the combustion and evaporation of fuel for all transport activity (excluding military transport), regardless of the sector, specified by sub-categories below.

Emissions from fuel sold to any air or marine vessel engaged in international transport (1 A 3 a i and 1 A 3 d i) should as far as possible be excluded from the totals and subtotals in this category and should be reported separately.""",
    )
    _c("1 A 3 a", "Civil Aviation")
    _c("1 A 3 a i", "International Aviation (International Bunkers)")
    _c("1 A 3 a ii", "Domestic Aviation")
    _c("1 A 3 b", "Road Transportation")
    _c("1 A 3 b i", "Cars")
    _c("1 A 3 b i 1", "Passenger cars with 3-way catalysts")
    _c("1 A 3 b i 2", "Passenger cars without 3-way catalysts")
    _c("1 A 3 b ii", "Light duty trucks")
    _c("1 A 3 b ii 1", "Light-duty trucks with 3-way catalysts")
    _c("1 A 3 b ii 2", "Light-duty trucks without 3-way catalysts")
    _c("1 A 3 b iii", "Heavy duty trucks and buses")
    _c("1 A 3 b iv", "Motorcycles")
    _c("1 A 3 b v", "Evaporative emissions from vehicles")
    _c("1 A 3 b vi", "Urea-based catalysts")
    _c("1 A 3 c", "Railways")
    _c("1 A 3 d", "Water-borne Navigation")
    _c("1 A 3 d i", "International water-borne navigation (International bunkers)")
    _c("1 A 3 d ii", "Domestic water-borne Navigation")
    _c("1 A 3 e", "Other Transportation")
    _c("1 A 3 e i", "Pipeline Transport")
    _c("1 A 3 e ii", "Off-road")
    _c("1 A 4 c iii", "Fishing (mobile combustion)")
    _c("1 A 5 a", "Non specified stationary")
    _c("1 A 5 b", "Non specified mobile")

    return cl


def gen_cs_ch3(**kwargs) -> "sdmx.model.common.ConceptScheme":
    """Generate a scheme of concepts included in equations in Chapter 3.

    The generated code list's URN ends with
    ``ConceptScheme=TDCI:CS_IPCC_2006_V2_CH3(…)``.

    .. todo:: Include concepts used as table dimensions.

    .. todo:: Include internationalized texts (names, descriptions) from the Arabic,
       Chinese, French, Russian, and/or Spanish versions of the documents.
    """
    from sdmx.model.common import Concept, ConceptScheme
    from sdmx.model.v21 import Annotation

    cs = ConceptScheme(
        id="CS_IPCC_2006_V2_CH3", name="Concepts appearing in equations", **kwargs
    )

    equation, page = "", ""

    def _c(id_, name=None, units=None, description=None):
        c = Concept(
            id=id_,
            name=name,
            description=f"First appears in Equation {equation} on p.{page}",
        )

        if units:
            c.annotations.append(Annotation(id="preferred-units", text=units))
        cs.append(c)

    # §3.2 Road transportation

    equation, page = "3.2.1", "3.12"
    _c(
        "EMI 1",
        "Emissions",
        "kg",
        """Variously "Emissions of CO₂" (Eq. 3.2.1), or of varying species (Eq. 3.2.3, 3.2.5)""",
    )
    _c("Fuel 1", "Fuel sold", "TJ")
    _c("EF 1", "Emission factor", "kg/TJ")
    _c(
        "a",
        "Type of fuel (e.g. petrol, diesel, natural gas, lpg)",
        None,
        "In Eq 3.2.6, 'j' is used for the same concept.",
    )

    equation, page = "3.2.2", "3.12"
    _c(
        "EMI 2",
        "CO₂ Emissions from urea-based additive in catalytic converters",
        "Gg CO₂",
    )
    _c(
        "Activity",
        "amount of urea-based additive consumed for use in catalytic converters",
        "Gg",
    )
    _c(
        "Purity",
        "the mass fraction (=percentage divided by 100) of urea in the urea-based additive",
    )

    # Eq. 3.2.3 —same concepts as 3.2.1

    equation, page = "3.2.4", "3.13"
    _c(
        "Fuel 2",
        "fuel consumed (as represented by fuel sold) for a given mobile source activity",
        "TJ",
    )
    _c(
        "b",
        "vehicle type",
        None,
        "In Eq 3.2.6, 'i' is used for the same concept (e.g., car, bus)",
    )
    _c(
        "c",
        "emission control technology (such as uncontrolled, catalytic converter, etc.)",
    )

    equation, page = "3.2.5", "3.15"
    _c("EF 2", "emission factor", "kg / km")
    _c(
        "Distance 1",
        "distance travelled during thermally stabilized engine operation phase for a given mobile source activity",
        "km",
    )
    _c("C", "emissions during warm-up phase (cold start)", "kg")
    _c(
        "d",
        "operating conditions (e.g. urban or rural road type, climate, or other environmental factors)",
    )

    equation, page = "3.2.6", "3.26"
    _c(
        "Estimated fuel",
        "total estimated fuel use estimated from distance travelled (VKT) data",
        "litre",
    )
    _c("Vehicles", "number of vehicles of type i and using fuel j on road type t")
    _c(
        "Distance 2",
        "annual kilometres travelled per vehicle of type i and using fuel j on road type t",
        "km",
    )
    _c("t", "type of road (e.g., urban, rural)")

    # §3.3 Off-road transportation
    # Eq. 3.3.1 —no additional concepts
    # Eq. 3.3.2 —no additional concepts

    equation, page = "3.3.3", "3.34"
    _c(
        "N",
        "source population",
        None,
        """In Eq. 3.4.3 this is given as 'number of locomotives of type i".""",
    )
    # Ditto below, all used in Eq. 3.4.3
    _c("H", "annual hours of use of vehicle i", "hour")
    _c("P", "average rated power of vehicle i", "kW")
    _c("LF", "typical load factor of vehicle i (fraction between 0 and 1)")
    _c("EF 3", "average emission factor for use of fuel j in vehicle i", "kg / kWh")

    # Eq. 3.3.4 —no additional concepts

    # §3.4 Railways
    # Eq. 3.4.1 —no additional concepts
    # Eq. 3.4.2 —no additional concepts

    equation, page = "3.4.3", "3.42"
    _c("i", "locomotive type and journey type")

    equation, page = "3.4.4", "3.43"
    _c("EF 4", "engine specific emission factor for locomotive of type i", "kg/TJ")
    _c("PWF", "pollutant weighting factor for locomotive of type i", "dimensionless")
    _c("EF 5", "default emission factor for diesel (applies to CH₄, N₂O)", "kg/TJ")

    # §3.6 Civil Aviation
    # Eq. 3.6.1 —no additional concepts

    equation, page = "3.6.2", "3.59"
    _c(
        "Emissions.LTO",
        "",
        None,
        """'LTO' is defined on p.3.56 as "Landing/Take-Off cycle".""",
    )
    _c(
        "Emissions.Cruise",
        "",
        None,
        """'Cruise' is defined on p.3.56 in contrast with 'LTO'.""",
    )

    equation, page = "3.6.3", "3.59"
    _c("Number of LTOs")
    _c("EF.LTO")

    equation, page = "3.6.4", "3.59"
    _c("Fuel consumption.LTO")
    _c("Fuel consumption per LTO")

    equation, page = "3.6.5", "3.59"
    _c("Total Fuel Consumption")
    _c("EF.Cruise")

    return cs


def gen_structures() -> None:
    """Create or update IPCC-maintained structural metadata."""
    from transport_data import STORE, org

    ma_args = dict(
        maintainer=org.get_agencies()[0],
        version="0.1.0",
        is_final=True,
        is_external_reference=False,
    )

    STORE.set(gen_cl_T311(**ma_args))
    STORE.set(gen_cs_ch3(**ma_args))



"""Utilities for LibreOffice."""

import logging
import subprocess
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pathlib

log = logging.getLogger(__name__)

#: Expected name of the LibreOffice command-line program.
PROGRAM = "soffice"

try:
    assert (
        subprocess.run([PROGRAM, "--version"], stdout=subprocess.DEVNULL).returncode
        == 0
    )
except Exception:
    #: :any:`True` if :data:`PROGRAM` is present on the system
    HAS_LIBREOFFICE = False
else:
    HAS_LIBREOFFICE = True


def to_pdf(path: "pathlib.Path") -> None:
    """Convert LibreOffice-compatible file at `path` to PDF."""
    if not HAS_LIBREOFFICE:
        log.info(f"{PROGRAM!s} not found; skip conversion of {path}")
        return

    subprocess.check_call(
        [PROGRAM, "--headless", "--convert-to", "pdf", "--outdir"]
        + [str(path.parent), str(path)]
    )



import pytest
from requests.exceptions import ConnectionError, ConnectTimeout

from transport_data.jrc import convert, fetch


@pytest.mark.xfail(
    raises=(
        NotImplementedError,  # Test is incomplete
        # Fetching data occasionally fails
        # TODO Use flaky mark instead
        ConnectionError,
        ConnectTimeout,
    )
)
def test_convert(geo="AT"):
    fetch(geo)
    convert(geo)



"""Intergovernmental Panel on Climate Change metadata provider."""

from transport_data.util.pluggy import hookimpl


@hookimpl
def get_agencies():
    """Return the IPCC :class:`.Agency`."""
    from sdmx.model import common

    a = common.Agency(
        id="IPCC",
        name="Intergovernmental Panel on Climate Change",
        description="https://www.ipcc.ch/",
    )
    return (a,)


@hookimpl
def provides():
    return (
        "Codelist=TDCI:CL_IPCC_2006_V2_T3.1.1",
        "ConceptScheme=TDCI:CS_IPCC_2006_V2_CH3",
    )



"""Utilities for working with pycountry_.

.. _pycountry: https://pypi.org/project/pycountry/
"""

from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Mapping

if TYPE_CHECKING:
    import gettext

    import pycountry.db

#: Fields in the :mod:`.pycountry` databases for which localizations are available.
LOCALIZABLE = {
    "name",
    "official_name",
}

#: Mapping from country name forms appearing in data to values recognized by
#: :meth:`pycountry.countries.lookup`.
NAME_MAP = {
    "mainland china": "CN",
    "russia": "RU",  # "Russian Federation"
    "taiwan, china": "TW",
    "turkey": "TR",  # "Türkiye"
}


@dataclass
class DatabaseInfo:
    """Information about a :mod:`pycountry` database."""

    #: Name of the variable in the :mod:`pycountry` namespace.
    name: str
    #: Matches :attr:`pycountry.db.Database.root_key`.
    root_key: str
    #: Fields which can be used as code IDs.
    id_fields: set[str]


#: Information on the databases of :mod:`pycountry`.
DATABASE_INFO = (
    DatabaseInfo("countries", "3166-1", {"alpha_2", "alpha_3", "numeric"}),
    DatabaseInfo("subdivisions", "3166-2", {"code"}),
    DatabaseInfo(
        "historic_countries", "3166-3", {"alpha_2", "alpha_3", "alpha_4", "numeric"}
    ),
    DatabaseInfo("currencies", "4217", {"alpha_3", "numeric"}),
    DatabaseInfo("languages", "639-3", {"alpha_2", "alpha_3"}),
    DatabaseInfo("language_families", "639-5", {"alpha_3"}),
    DatabaseInfo("scripts", "15924", {"alpha_4", "numeric"}),
)


@lru_cache
def get_database(standard_number: str) -> tuple["pycountry.db.Database", DatabaseInfo]:
    """Retrieve a :mod:`pycountry` database given its ISO `standard_number`."""
    import pycountry

    for info in DATABASE_INFO:
        if info.root_key == standard_number:
            return getattr(pycountry, info.name), info

    raise ValueError(standard_number)


@lru_cache
def load_translations(domain: str) -> Mapping[str, "gettext.NullTranslations"]:
    """Load all available :mod:`pycountry` translations for `domain`."""
    from gettext import translation

    from pycountry import LOCALES_DIR

    # Iterate over all subdirectories of the pycountry locale dir
    result = {}
    for lang in map(lambda d: d.name, Path(LOCALES_DIR).iterdir()):
        try:
            result[lang] = translation(domain, LOCALES_DIR, languages=[lang])
        except FileNotFoundError:
            pass  # No translations for this (domain, lang)

    return result



"""Eurostat (ESTAT) provider.

Eurostat is used as to demonstrate interaction with a data provider that offers:

- A web service that conforms to the SDMX REST API standard.
- Data and metadata in the SDMX-ML formatl through this API.

In this case, TDC tools can retrieve the structure information and data directly, and no
provider-specific conversion code is needed.
"""

import click
import sdmx

from transport_data import STORE

# General functions


def list_flows():
    """Return a list of data flows."""
    # TODO use the API to retrieve a list of transport-related data flows
    return ["TRAN_HV_MS_FRMOD"]


def get(dataflow_id: str):
    """Retrieve the ESTAT structure and data for the data flow with ID `dataflow_id`."""
    client = sdmx.Client("ESTAT")

    # Retrieve structural information as an sdmx.StructureMessage
    sm = client.dataflow(dataflow_id)

    # Write each of the structure objects received to a separate file
    STORE.update_from(sm)

    # Retrieve the data itself
    dm = client.data(dataflow_id)

    # Extract a single data set from the data message
    ds = dm.data[0]
    # Assign an explicit reference to the data set. The ESTAT response comes back with
    # an "anonymous" dataflow.
    ds.described_by = sm.dataflow[dataflow_id]
    ds.structured_by = ds.described_by.structure

    # Write to file
    path = STORE.set(ds)
    print(f"Retrieved {path}")
    print("          and associated structures")

    return path


# Command-line interface


@click.group("estat", help=__doc__)
def main():
    pass


@main.command("fetch")
def fetch_cmd():
    """Retrieve data."""
    for dataflow_id in list_flows():
        print(f"Retrieve data flow {dataflow_id!r} and structure")
        get(dataflow_id)



"""Utilities for :mod:`pluggy`."""

from importlib import import_module

import pluggy

from . import hooks

hookimpl = pluggy.HookimplMarker("transport_data")


pm = pluggy.PluginManager("transport_data")
pm.add_hookspecs(hooks)


def register_internal(*submodules: str) -> None:
    """Register hook implementations from `submodules` of :mod:`transport_data`."""
    for submodule in submodules:
        pm.register(import_module(f"transport_data.{submodule}"))



"""Other data providers."""

from transport_data.util.pluggy import hookimpl


@hookimpl
def get_agencies():
    from sdmx.model.common import Agency, Contact

    a0 = Agency(
        id="GAIKINDO",
        name={
            "en": "The Association of Indonesia Automobile Industries",
            "id": "Gabungan Industri Kendaraan Bermotor Indonesia",
        },
        contact=[
            Contact(
                uri=[
                    "https://www.gaikindo.or.id",
                    "https://www.gaikindo.or.id/indonesian-automobile-industry-data/",
                ]
            )
        ],
    )
    a1 = Agency(
        id="TAIA",
        name={
            "en": "The Thai Automotive Industry Association",
            "th": "สมาคมอุตสาหกรรมยานยนต์ไทย",
        },
        contact=[Contact(uri=["https://taia.or.th", "https://taia.or.th/en/"])],
    )

    return (a0, a1)



"""CLI for :mod:`.iso`.

.. runblock:: console

   $ tdc iso --help

"""

import click


@click.group(name="iso")
def main():
    """International Organization for Standardization (ISO) provider."""


@main.command
def refresh():  # pragma: no cover
    """Regenerate the ISO code lists."""
    from transport_data.iso import generate_all

    generate_all()



"Edit SDMX interactively."

from abc import ABC, abstractmethod
from math import ceil, log10
from typing import cast

import click
import prompt_toolkit
from sdmx.model import common, v21

#: Kinds of :class:`.MaintainableArtefacts` that can be edited.
CLASSES = (
    common.AgencyScheme,  # 1
    common.Codelist,  # 2
    common.ConceptScheme,  # 3
    v21.DataflowDefinition,  # 4
    v21.DataStructureDefinition,  # 5
)

HELP_TEXT = "Control-C: exit"


class EditorState:
    """References to artefacts being created or edited."""

    ia: common.IdentifiableArtefact | None = None
    na: common.NameableArtefact | None = None
    ma: common.MaintainableArtefact | None = None

    view: "View | None" = None


class Editor(prompt_toolkit.Application):
    """:class:`prompt_toolkit.Application` for editing SDMX maintainable artefacts.

    The editor has a simple UI with an output pane and a configurable prompts. These are
    set by :class:`.View` subclasses, which each implement :meth:`.View.accept` to
    accept and handle user input. Transitions between views are mainly described by
    :data:`FLOW`.
    """

    current: EditorState

    def __init__(self, *args, **kwargs) -> None:
        from prompt_toolkit.key_binding import KeyBindings
        from prompt_toolkit.layout.containers import HSplit
        from prompt_toolkit.layout.layout import Layout
        from prompt_toolkit.widgets import HorizontalLine, Label, TextArea

        # Build the visual layout
        self.output_field = Label(text=HELP_TEXT, dont_extend_height=False)
        self.input_field = TextArea(
            height=1,
            prompt=">>> ",
            style="class:input-field",
            multiline=False,
            wrap_lines=False,
            accept_handler=self.accept,
        )
        container = HSplit(
            [
                self.output_field,
                HorizontalLine(),
                self.input_field,
                Label(text=HELP_TEXT),
            ]
        )
        layout = Layout(container, focused_element=self.input_field)

        # Connect keybindings
        kb = KeyBindings()
        kb.add("c-c")(self._exit)

        # State for other classes in this module
        self.current = EditorState()

        # Construct the Application
        super().__init__(key_bindings=kb, layout=layout, full_screen=True)

        # Start with the first View in FLOW
        self.next()

    @staticmethod
    def _exit(event):
        """Control-C handler."""
        event.app.exit()

    def next(self, view: type["View"] | None = None) -> None:
        """Identify the next :class:`.View` `cls`; display it and its prompt.

        If `view` is :any:`None` (the default), :data:`FLOW` is used to identify the
        class.
        """

        # Look up the next view using FLOW
        if self.current.view is None:
            cls: type["View"] = cast(type["View"], FLOW[None])
            assert isinstance(cls, type)
        elif view is None:
            # No view
            class_or_map = FLOW[type(self.current.view)]
            obj = self.current.ia
            if class_or_map is None:
                # Exit the application
                self.exit()
                return
            elif isinstance(class_or_map, type):
                cls = class_or_map
            else:
                try:
                    cls = [c for t, c in class_or_map.items() if isinstance(obj, t)][0]
                except IndexError:  # pragma: no cover
                    raise RuntimeError(
                        f"No view to follow {type(self.current.view)} among "
                        f"{class_or_map}"
                    )
        else:
            cls = view

        # Create the next view
        self.current.view = cls(self)

    def accept(self, buff) -> bool:
        """Handle input to the input_field."""
        # Call the accept() method on the current View
        assert self.current.view
        cls = self.current.view.accept(self.input_field.text)
        # Pass the class of the next View (if any was returned) to .next()
        self.next(cls)

        return True

    def set_prompt(self, value: str, default: str = "") -> None:
        """Set the input prompt and default text."""
        from prompt_toolkit.formatted_text import FormattedText
        from prompt_toolkit.layout.processors import BeforeInput

        c = self.input_field.control
        assert c.input_processors and isinstance(c.input_processors[-1], BeforeInput)
        c.input_processors[-1].text = FormattedText([("bold", value + " ")])

        self.input_field.text = default
        self.input_field.buffer.cursor_position = len(default)


class View(ABC):
    """Representation of a :class:`.Editor` view."""

    #: Reference to the Editor.
    app: Editor
    #: Reference to the EditorState.
    current: EditorState

    #: Text to be displayed in the output text area.
    text: str
    #: Prompt text to be displayed in the input text area.
    prompt: str
    #: Default input.
    default: str = ""

    def __init__(self, app) -> None:
        """Create and display the view.

        Subclasses **may** override this method, but **must** call
        :py:`super().__init__(app)`.
        """
        self.app = app
        self.current = app.current

        # Display the view
        app.output_field.formatted_text_control.text = self.text
        app.set_prompt(self.prompt, self.default)

    @abstractmethod
    def accept(self, text: str) -> type["View"] | None:
        """Handle user input.

        Subclasses **must** implement this method. An implementation **may** return a
        :class:`.View` subclass (not instance) to indicate the next View to be
        displayed; if it returns :any:`None`, then :data:`NEXT_VIEW` is used to identify
        the next View.
        """


# Concrete View subclasses, in alphabetical order


class ComponentListEdit(View):
    """Base class for editing a :class:`.ComponentList` on ."""

    _cl_name = ""

    def __init__(self, app):
        # Reference to the ComponentList
        self._component_list = getattr(app.current.ma, self._cl_name + "s")
        self.prompt = f"Enter the ID of the next {self._cl_name}, or [enter] to stop:"

        super().__init__(app)

    @property
    def text(self):
        return dsd_text(self.current.ma)

    def accept(self, text):
        if text:
            self._component_list.getdefault(id=text)
            return type(self)  # Add another component
        else:
            return None  # Next view according to FLOW


class DFDStructureURN(View):
    """Set :attr:`.BaseDataflowDefinition.structure` based on the URN of a DSD."""

    @property
    def text(self):
        lines = [
            f"Creating a new data flow definition: {self.current.ma!r}\n",
            "Enter the full or partial URN for the data structure definition (DSD).",
        ]
        return "\n".join(lines)

    prompt = "Enter the URN:"

    def accept(self, text):
        from transport_data import STORE

        try:
            dsd = STORE.get(text)
        except KeyError:  # pragma: no cover  # TODO Extend tests to cover this error
            return DFDStructureURN
        self.current.ma.structure = dsd


class DSDAddDimension(ComponentListEdit):
    _cl_name = "dimension"


class DSDAddMeasure(ComponentListEdit):
    _cl_name = "measure"


class DSDAddAttribute(ComponentListEdit):
    _cl_name = "attribute"


class IA_ID(View):
    """:attr:`.IdentifiableArtefact.id`."""

    @property
    def text(self):
        return f"Creating a new artefact: {self.current.ia!r}"

    prompt = "Enter its ID:"

    def accept(self, text):
        self.current.ia.id = text


class ItemSchemeEdit(View):
    """Edit a :class:`.ItemScheme`."""

    def __init__(self, app):
        current = app.current
        if isinstance(current.na, current.ma._Item):
            # Adding an item to an ItemScheme. Store and clear.
            current.ma.append(current.na)
            current.ia = current.na = current.ma

        super().__init__(app)

    @property
    def text(self):
        ma = self.current.ma
        lines = [
            f"Creating a new {type(ma).__name__}: {ma!r}\n",
            f"containing {len(ma)} {ma._Item.__name__}(s):\n",
        ]

        # List the items
        width = ceil(log10(len(ma) + 1))
        for i, item in enumerate(ma.items.values(), start=1):
            lines.append(f"{i:{width}}. {item!r}")

        return "\n".join(lines)

    prompt = "Enter a number to edit; [n]ew item; or [enter] to stop:"

    def accept(self, text):
        if text == "n":  # New Item
            self.current.ia = self.current.na = self.current.ma._Item()
        elif text == "":  # Done adding Items; save
            return MA_Save
        else:
            try:
                # Interpret `text` as the number of an Item
                i = int(text)
                self.current.ia = self.current.na = list(
                    self.current.ma.items.values()
                )[i - 1]
            except (
                ValueError,  # Not an integer
                IndexError,  # Not an integer that is the index of an existing Item
            ):
                return type(self)


class MA_Class(View):
    """Class for a :class:`.MaintainableArtefact`."""

    prompt = "Enter a number from the above list:"

    @property
    def text(self):
        lines = [
            "Creating a new maintainable artefact.\n",
            "Choose a class from among:\n",
        ]
        for i, cls in enumerate(CLASSES, start=1):
            lines.append(f"{i:2}. {cls.__module__}.{cls.__name__}")
        return "\n".join(lines)

    def accept(self, text):
        try:
            ma_class = CLASSES[int(text) - 1]
            self.current.ia = self.current.na = self.current.ma = (
                ma_class()
            )  # Create the object to be constructed
        except (
            ValueError,  # Non-integer input
            IndexError,  # A value that is not in the enumeration of `choices`
        ):
            return type(self)  # Repeat the same view


class MA_Maintainer(View):
    """:attr:`.MaintainableArtefact.maintainer`."""

    @property
    def text(self):
        return f"Creating a new artefact: {self.current.ma!r}"

    prompt = "Enter the maintainer ID:"
    default = "TDCI"

    def accept(self, text):
        self.current.ma.maintainer = common.Agency(id=text)


class MA_Save(View):
    """Save a completed :class:`.MaintainableArtefact`."""

    @property
    def text(self):
        return f"Creating a new artefact: {self.current.ma!r}"

    prompt = "Save? [y/n]"

    def accept(self, text):
        if text != "y":
            return

        from transport_data import STORE

        STORE.set(self.current.ma)


class NA_Name(View):
    """:attr:`.NameableArtefact.name`."""

    prompt = "Enter its name or [enter] to skip:"

    @property
    def text(self):
        return f"Creating a new artefact: {self.current.na!r}"

    def accept(self, text):
        if text:
            self.current.na.name = text


class VA_Version(View):
    """:attr:`.VersionableArtefact.version`."""

    @property
    def text(self):
        return f"Creating a new artefact: {self.current.ma!r}"

    prompt = "Enter its version:"
    default = "1.0.0"

    def accept(self, text):
        self.current.ma.version = text


#: Mapping from Views to the following View to be displayed.
#:
#: If the value is a :class:`dict`, it maps from classes for :attr:`.EditorState.ia` to
#: Views.
FLOW: dict[type[View] | None, None | type[View] | dict[type, type[View]]] = {
    None: MA_Class,
    MA_Class: MA_Maintainer,
    MA_Maintainer: IA_ID,
    IA_ID: NA_Name,
    NA_Name: {
        common.Item: ItemSchemeEdit,
        common.VersionableArtefact: VA_Version,
    },
    VA_Version: {
        common.BaseDataStructureDefinition: DSDAddDimension,
        common.BaseDataflow: DFDStructureURN,
        common.ItemScheme: ItemSchemeEdit,
    },
    DFDStructureURN: MA_Save,
    DSDAddDimension: DSDAddMeasure,
    DSDAddMeasure: DSDAddAttribute,
    DSDAddAttribute: MA_Save,
    ItemSchemeEdit: IA_ID,
    MA_Save: None,
}


def dsd_text(dsd: v21.DataStructureDefinition) -> str:
    """Generate display text about a DSD."""
    lines = [f"Creating a new data structure definition: {dsd!r}"]

    for label, cl in (
        ("dimension", dsd.dimensions),
        ("measure", dsd.measures),
        ("attribute", dsd.attributes),
    ):
        lines.extend(["", f"with {len(cl)} {label}(s):"])
        for i, component in enumerate(cl.components, start=1):
            lines.append(f"{i:2}. {component!r}")

    return "\n".join(lines)


@click.command("edit", help=__doc__)
def main():  # pragma: no cover
    Editor().run()



import logging
from functools import partial
from pathlib import Path

import pandas as pd
from sdmx.model import common, v21

from transport_data.util.google import get_service

log = logging.getLogger(__name__)

# If modifying these scopes, delete the file token.json.
SCOPES = [
    "https://www.googleapis.com/auth/drive.readonly",
    "https://www.googleapis.com/auth/drive.metadata.readonly",
    # "https://www.googleapis.com/auth/spreadsheets.readonly",
]

# The file ID of the Google Drive file.
FILE_ID = "1uMuNG9rTGO52Vuuq6skyqmkH9U5yv1iSJDJYjH64MJM"


def convert(path: Path) -> "v21.DataSet":
    return read_workbook(path)


def read_workbook(path: Path) -> "v21.DataSet":
    """Read a metadata set from the workbook at `path`."""
    ef = pd.ExcelFile(path)

    return read_worksheet_rtr(ef)


def get_dsd_rtr() -> "v21.DataStructureDefinition":
    cs_measure = common.ConceptScheme(id="MEASURE")
    cs_measure.setdefault(
        id="RTR",
        name="Rapid Transit to Resident Ratio (RTR)",
        description="km of Rapid Transit / Million Urban Residents (in cities with populations above 500,000)",
    )

    result = v21.DataStructureDefinition(id="RTR")

    result.dimensions.getdefault(id="GEO")
    result.dimensions.getdefault(id="TYPE")
    result.dimensions.getdefault(id="TIME_PERIOD")

    result.measures.getdefault(id="OBS_VALUE", concept_identity=cs_measure["RTR"])

    return result


def read_worksheet_rtr(ef: "pd.ExcelFile", sheet_name="Country RTR") -> "v21.DataSets":
    """Read the “Country RTR” sheet of the RTDB."""
    from pycountry import countries

    from transport_data.util.pycountry import NAME_MAP
    from transport_data.util.sdmx import make_obs

    def _geo_alpha_2(value: str) -> str:
        try:
            return countries.lookup(NAME_MAP.get(value.lower(), value)).alpha_2
        except LookupError:
            if value == "Grand TOTAL":
                return "_T"
            else:
                raise

    # Read and transform the worksheet
    df = (
        ef.parse(sheet_name=sheet_name)
        .rename(columns={"RTR": "GEO", "Unnamed: 2": "TYPE"})
        .drop("Region", axis=1)
        .assign(
            GEO=lambda df: df["GEO"].ffill().apply(_geo_alpha_2),
            TYPE=lambda df: df["TYPE"].str.replace("Total", "_T"),
        )
        .set_index(["GEO", "TYPE"])
        .melt(var_name="TIME_PERIOD", value_name="OBS_VALUE", ignore_index=False)
        .reset_index()
    )

    # Obtain the data structure definition
    dsd = get_dsd_rtr()

    # Convert data to SDMX
    return v21.DataSet(
        structured_by=dsd, obs=df.apply(partial(make_obs, dsd=dsd), axis=1).to_list()
    )


FETCH_MIME_TYPE = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"


def fetch() -> Path:
    """Fetch the RTDB file.

    This function uses the Google Drive API v3 to export the upstream database
    """
    from datetime import datetime, timezone

    from platformdirs import user_cache_path

    cache_dir = user_cache_path("transport-data").joinpath("itdp")
    cache_dir.mkdir(parents=True, exist_ok=True)
    cache_path = cache_dir.joinpath("rtdp.xlsx")

    # Local modified time
    tz = datetime.now(timezone.utc).astimezone().tzinfo
    try:
        mtime_cache = datetime.fromtimestamp(cache_path.stat().st_mtime, tz=tz)
    except FileNotFoundError:
        mtime_cache = datetime.fromtimestamp(0, tz=tz)

    # Call the Drive API
    try:
        drive = get_service("drive", "v3", scopes=SCOPES).files()
        result = drive.get(fileId=FILE_ID, fields="modifiedTime").execute()

        # Last modified time of the remote file
        mtime_remote = datetime.fromisoformat(result["modifiedTime"])
    except RuntimeError as e:
        if cache_path.exists():
            # Continue with returning cache_path, even though it may be stale
            log.warning(
                f"Could not check remote file modification time; {cache_path} may be "
                "stale"
            )
            log.debug(e)
        else:
            raise RuntimeError("Could not fetch remote file") from e
    else:
        if not cache_path.exists() or mtime_cache < mtime_remote:
            # Export the file to .xlsx
            request = drive.export(fileId=FILE_ID, mimeType=FETCH_MIME_TYPE)

            # Write to the cache path
            with open(cache_path, "wb") as f:
                f.write(request.execute())

    return cache_path



"""CLI for :mod:`.itdp`.

.. runblock:: console

   $ tdc itdp --help

.. runblock:: console

   $ tdc itdp fetch --help
"""

import click


@click.group("itdp")
def main():
    """Institute for Transport and Development Policy (ITDP).

    Fetch and convert data from the Rapid Transit Database.
    """
    from .rtdb import convert, fetch

    result = convert(fetch())
    print(result)



"""Utilities."""

from itertools import chain
from urllib.parse import urlparse


def list_urns() -> list[str]:
    """List all URNs of SDMX artefacts provided by :mod:`tranport_data` and plugins."""
    from transport_data.util.pluggy import pm

    return list(chain(*pm.hook.provides()))


def uline(text: str, char: str = "=") -> str:
    """Underline `text` with `char`."""
    return f"{text}\n{char * len(text)}"


def url_hostname(value: str) -> str:
    """If `value` contains a URL, return its hostname."""
    try:
        return urlparse(value).hostname or ""
    except Exception:
        return value



"""Generate reports based on data and metadata."""

from abc import ABC, abstractmethod
from pathlib import Path


class Report(ABC):
    """Abstract class for reports.

    Subclasses:

    - **must** implement :meth:`render`.
    - **may** use :func:`dataclasses.dataclass` to declare additional attributes and an
      :py:`__init__()` method that accepts and stores them.
    """

    #: Name of a Jinja2 template used by the report; see :meth:`render_jinja_template`.
    template_name: str

    @abstractmethod
    def render(self) -> str | bytes:
        """Render the report (generate its contents) and return as str or bytes.

        The content may be in any format: plain text, HTML, binary file content, etc.
        """

    def render_jinja_template(self, *args, **kwargs) -> str:
        """Retrieve the Jinja2 :attr:`template_name` and call its render method."""
        from transport_data.util import jinja2

        env, common = jinja2.get_env()

        return env.get_template(self.template_name).render(*args, **kwargs, **common)

    def rst2odt(self, content: str) -> bytes:
        """Convert `content` from reStructuredText to OpenDocument Text (ODT).

        Returns
        -------
        bytes
            The ODT (ZIP) archive.
        """
        from docutils.core import publish_string

        from transport_data.util import docutils, jinja2

        env, _ = jinja2.get_env()

        ss_path = Path(env.get_template("odt-styles.xml").filename)
        settings = {"create_links": True, "stylesheet": str(ss_path)}

        # Convert reStructuredText → ODF ODT using docutils
        return publish_string(
            writer=docutils.ODTWriter(), source=content, settings_overrides=settings
        )

    def write_file(self, path: Path, **kwargs) -> None:
        """:meth:`render` the report and write to `path`."""
        content = self.render()
        if isinstance(content, str):
            path.write_text(content, **kwargs)
        else:
            path.write_bytes(content, **kwargs)



import re

import pytest
from prompt_toolkit.input.ansi_escape_sequences import REVERSE_ANSI_SEQUENCES
from prompt_toolkit.keys import Keys

from transport_data.cli.interactive import Editor
from transport_data.testing import ember_dfd


@pytest.mark.parametrize(
    "command",
    (
        ("--help",),
        ("ato", "--help"),
        ("ato", "fetch", "--all"),
        ("check", "--help"),
        ("config", "--help"),
        ("estat", "--help"),
        ("estat", "fetch", "--help"),
        ("iamc", "--help"),
        ("iamc"),
        ("jrc", "--help"),
        ("jrc", "fetch", "--all"),
        ("org", "--help"),
        ("org", "--version=0.0.0"),
        ("proto", "--help"),
        ("store", "--help"),
    ),
)
def test_cli(tdc_cli, command):
    tdc_cli.invoke(command)


CHECK_ARGS = [
    "--structure=dataflow",
    "--structure-id=FOO",
    "--action=I",
    "--sheets=sheet_2",
    "Dataflow=TDCI:EMBER_001(1.0.0)",
]


def test_check0(tdc_cli, test_data_path, tmp_store):
    """Check a successful read of a .xlsx file."""
    ember_dfd(tmp_store)

    path = test_data_path.joinpath("read-csv-2.xlsx")
    result = tdc_cli.invoke(["check"] + CHECK_ARGS + [str(path)])

    # Command runs without error
    assert 0 == result.exit_code, result.output

    # Data was located, converted to CSV, read, and summary displayed
    assert re.fullmatch(
        r"""
File: .*read-csv-2.xlsx
Sheet: sheet_2

1 data set\(s\) in: Dataflow.*=TDCI:EMBER_001\(1.0.0\)

Data set 0: action=ActionType.information
120 observations
""",
        result.output,
    )


@pytest.mark.parametrize(
    "args, exit_code, text",
    (
        (["X"], 1, "line 1, field 1"),  # Omit all options and URN
        (CHECK_ARGS[:1] + ["X"], 1, "line 1, field 2"),  # Only --structure=
        (CHECK_ARGS[:2] + ["X"], 1, "'X' could not be loaded"),  # Invalid structure URN
        (CHECK_ARGS + ["-v"], 0, "..."),  # Show pd.DataFrame abbreviated string repr
        (CHECK_ARGS + ["-vv"], 0, ""),  # Show pd.DataFrame full string repr
    ),
)
def test_check1(tdc_cli, test_data_path, tmp_store, args, exit_code, text):
    """Check various other argument combinations."""
    ember_dfd(tmp_store)

    path = test_data_path.joinpath("read-csv-1.csv")
    result = tdc_cli.invoke(["check"] + args + [str(path)])

    # Command gives the expected exit code
    assert exit_code == result.exit_code, result.output

    # Expected text is contained in the output
    if text:
        assert text in result.output


def test_check2(tdc_cli, tmp_path):
    path = tmp_path.joinpath("foo.txt")
    path.touch()
    result = tdc_cli.invoke(["check", "X", str(path)])

    assert 2 == result.exit_code, result.output
    assert "Unsupported file extension" in result.output


def run_script(lines: list[str]) -> None:
    """Create a contained instance of :class:`.Editor` and feed it `lines`."""
    from prompt_toolkit.application import create_app_session
    from prompt_toolkit.input import create_pipe_input
    from prompt_toolkit.output import DummyOutput

    # Create an input pipe
    with create_pipe_input() as pipe_input:
        # Feed the script into the pipe
        for line in lines:
            pipe_input.send_text(line + "\r")

        # Create an app session; run the app within the session
        with create_app_session(input=pipe_input, output=DummyOutput()):
            Editor().run()


def _backspace(text: str) -> str:
    """Generate ANSI sequences for backspacing over `text`."""
    return len(text) * REVERSE_ANSI_SEQUENCES[Keys.ControlH]


#: ‘Script’ of CLI input to produce a code list.
SCRIPT_2 = [
    "99",  # Class: 99 → doesn't exist, return to the same meeting
    "2",  # Class: 2 = Codelist
    _backspace("TDCI") + "TEST",  # Maintainer ID
    "CL_CLI_EDIT",  # Artefact ID
    "",  # Artefact name
    "",  # Artefact version (accept default 1.0.0)
    "n",  # New Item
    "FOO",  # Item ID
    "Foo",  # Item name
    "n",
    "BAR",  # Item ID
    "Bar",
    "n",
    "BAZ",  # Item ID
    "",  # Item name
    "10",  # Edit 10th created item (doesn't exist, return to the same view)
    "2",  # Edit 2nd created item (BAR)
    "QUX",  # New Item ID
    "",  # Keep existing name
    "",  # Continue
    "y",  # Save
]


@pytest.mark.timeout(1)
def test_edit2(tmp_store) -> None:
    # CLI runs and accepts the input without error
    run_script(SCRIPT_2)

    # A code list is generated in the tmp_store
    cl = tmp_store.get("Codelist=TEST:CL_CLI_EDIT(1.0.0)")

    # The code list has 3 items
    assert {"FOO", "QUX", "BAZ"} == set(cl.items.keys())
    # The ID but not name of the second item was edited
    assert "Bar" == cl["QUX"].name["en"]


#: ‘Script’ of CLI input to produce a data flow definition.
SCRIPT_4 = [
    "4",  # Class: 4 = DataflowDefinition
    _backspace("TDCI") + "TEST",  # Maintainer ID
    "DF_CLI_EDIT",  # Artefact ID
    "",  # Artefact name
    "",  # Artefact version (accept default 1.0.0)
    "DataStructureDefinition=TEST:MASS(1.0.0)",  # URN of the DSD
    "y",  # Save
]


@pytest.mark.timeout(1)
@pytest.mark.usefixtures("sdmx_structures")
def test_edit4(tmp_store) -> None:
    # CLI runs and accepts the input without error
    run_script(SCRIPT_4)

    # A DFD is generated in the tmp_store
    tmp_store.get("Dataflow=TEST:DF_CLI_EDIT(1.0.0)")


#: ‘Script’ of CLI input to produce a data structure definition.
SCRIPT_5 = [
    "5",  # Class: 5 = DataStructureDefinition
    _backspace("TDCI") + "TEST",  # Maintainer ID
    "DS_CLI_EDIT",  # Artefact ID
    "",  # Artefact name
    "",  # Artefact version (accept default 1.0.0)
    "FOO",  # Dimension
    "BAR",
    "BAZ",
    "",  # Continue
    "OBS_VALUE",  # Measure
    "",
    "UNIT_MEASURE",  # Attribute
    "OBS_STATUS",
    "",
    "y",  # Save
]


@pytest.mark.timeout(1)
def test_edit5(tmp_store) -> None:
    # CLI runs and accepts the input without error
    run_script([REVERSE_ANSI_SEQUENCES[Keys.ControlC]])
    # Nothing was saved because ControlC was given
    with pytest.raises(KeyError):
        tmp_store.get("DataStructureDefinition=TEST:DS_CLI_EDIT(1.0.0)")

    # CLI runs and accepts the input without error
    run_script(SCRIPT_5[:-1] + [""])
    # Nothing was saved, because "y" was not given at the final, "Save" view
    with pytest.raises(KeyError):
        tmp_store.get("DataStructureDefinition=TEST:DS_CLI_EDIT(1.0.0)")

    # CLI runs and accepts the input without error
    run_script(SCRIPT_5)

    # A DSD is generated in the tmp_store
    dsd = tmp_store.get("DataStructureDefinition=TEST:DS_CLI_EDIT(1.0.0)")
    # It has contents according to the CLI inputs
    assert 3 == len(dsd.dimensions)
    assert 1 == len(dsd.measures)
    assert 2 == len(dsd.attributes)



"""Utilities for :mod:`click`."""

import click

PARAM = {"version": click.Option(["--version"])}


def common_params(names: str) -> list:
    """Return parameters from `PARAM`."""
    return list(PARAM[k] for k in names.split())



"""Utilities for :mod:`responses`."""

from typing import TYPE_CHECKING, Any

import responses

if TYPE_CHECKING:
    from responses import BaseResponse, PreparedRequest


class RepeatRegistry(responses.FirstMatchRegistry):
    """Response registry that repeats the same response for each URL infinitely."""

    def find(
        self, request: "PreparedRequest"
    ) -> tuple["BaseResponse | None", list[str]]:
        found = None
        found_match = None
        match_failed_reasons = []
        for i, response in enumerate(self.registered):
            match_result, reason = response.matches(request)
            if match_result:
                if found is None:
                    found = i
                    found_match = response
                else:
                    # Parent class uses self.registered.pop() below
                    if self.registered[found].call_count > 0:
                        found_match = response
                        break
                    # Multiple matches found → return the first response
                    return self.registered[found], match_failed_reasons
            else:
                match_failed_reasons.append(reason)
        return found_match, match_failed_reasons


class RequestsMock(responses.RequestsMock):
    """RequestsMock that does not reset on context manager exit."""

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        success = type is None
        self.stop(allow_assert=success)



from transport_data.org import get_agencyscheme, refresh


def test_get_agencyscheme() -> None:
    as_ = get_agencyscheme()

    # Number of agencies associated with code in the transport_data repo
    assert 11 == len(as_)


def test_refresh() -> None:
    refresh()



import pytest
from ckanapi.errors import (
    CKANAPIError,
    # NotAuthorized,
    ValidationError,
)
from pytest import param

from transport_data.util.ckan import (
    Client,
    Group,
    Organization,
    Package,
    Resource,
    get_class,
)

_500 = pytest.mark.xfail(raises=CKANAPIError, reason="500 Internal Server Error")
_Incomplete = pytest.mark.xfail(
    raises=ValidationError, reason="Incomplete test; needs args"
)
_NotImplemented = pytest.mark.xfail(
    raises=CKANAPIError, reason="Not implemented by .org.ckan.DEV"
)
# _NotAuthorized = pytest.mark.xfail(raises=NotAuthorized, reason="Needs API key")
_NotAuthorized = pytest.mark.skip(reason="Needs API key")


class TestOrganization:
    @pytest.fixture
    def obj(self, test_data_path) -> Organization:
        p = Package.from_file(test_data_path.joinpath("ckan", "package.json"))
        return Organization(name="test-org", packages=[p, p])

    def test_repr(self, obj: Organization) -> None:
        assert "<CKAN Organization 'test-org'" in repr(obj)
        assert "\n  <CKAN Package '2023-production" in repr(obj)


class TestPackage:
    """These also function as tests of :class:`.ModelProxy`."""

    @pytest.fixture
    def obj(self, test_data_path) -> Package:
        """:class:`.Package` instance from a test specimen file."""
        return Package.from_file(test_data_path.joinpath("ckan", "package.json"))

    def test_asdict(self, obj) -> None:
        obj.asdict()

    def test_get_item(self, obj) -> None:
        g = obj.get_item("groups", 0)
        assert isinstance(g, Group)

        r = obj.get_item("resources", 0)
        assert isinstance(r, Resource)

    def test_init_name_only(self) -> None:
        p = Package(name="foo-bar")
        assert "<CKAN Package 'foo-bar' | 0 fields | id=(none)>" == repr(p)

    def test_len(self, obj) -> None:
        assert 47 == len(obj)

    def test_update(self, obj) -> None:
        with pytest.raises(ValueError):
            obj.update({"id": "foo"})
        with pytest.raises(ValueError):
            obj.update({"name": "foo"})

    def test_repr(self, obj) -> None:
        assert "46 fields" in repr(obj)


class TestClient:
    @pytest.fixture
    def c(self) -> "Client":
        # FIXME Use a mock API
        from transport_data.org.ckan import DEV

        return DEV

    # Tests of explicit methods
    def test_package_list(self, c) -> None:
        result = c.package_list()
        assert 1 <= len(result)

    def test_package_show(self, c) -> None:
        packages = c.package_list(limit=1, max=1)
        p = c.package_show(packages[0])
        assert isinstance(p, Package)
        assert p.id is not None

    # Generic tests of other HTTP GET API endpoints via call_action
    @pytest.mark.parametrize(
        "action, args",
        [
            param("am_following_dataset", {}, marks=_Incomplete),
            param("am_following_group", {}, marks=_Incomplete),
            param("am_following_organization", {}, marks=_NotImplemented),
            param("am_following_user", {}, marks=_Incomplete),
            param("api_token_list", {}, marks=_Incomplete),
            param("config_option_list", {}, marks=_NotAuthorized),
            param("config_option_show", {}, marks=_NotAuthorized),
            ("current_package_list_with_resources", {}),
            param("dataset_followee_count", {}, marks=_Incomplete),
            param("dataset_followee_list", {}, marks=_NotAuthorized),
            param("dataset_follower_count", {}, marks=_Incomplete),
            param("dataset_follower_list", {}, marks=_NotAuthorized),
            param("followee_count", {}, marks=_Incomplete),
            param("followee_list", {}, marks=_Incomplete),
            param("format_autocomplete", {}, marks=_Incomplete),
            param("get_site_user", {}, marks=_NotAuthorized),
            param("group_autocomplete", {}, marks=_NotImplemented),
            param("group_followee_count", {}, marks=_Incomplete),
            param("group_followee_list", {}, marks=_NotAuthorized),
            param("group_follower_count", {}, marks=_Incomplete),
            param("group_follower_list", {}, marks=_NotAuthorized),
            ("group_list_authz", {}),
            ("group_list", {}),
            param("group_package_show", {}, marks=_Incomplete),
            param("group_show", {}, marks=_Incomplete),
            param("help_show", {}, marks=_Incomplete),
            param("job_list", {}, marks=_NotAuthorized),
            param("job_show", {}, marks=_NotAuthorized),
            ("license_list", {}),
            param("member_list", {}, marks=_Incomplete),
            ("member_roles_list", {}),
            param("organization_autocomplete", {}, marks=_500),
            param("organization_followee_count", {}, marks=_Incomplete),
            param("organization_followee_list", {}, marks=_NotAuthorized),
            param("organization_follower_count", {}, marks=_Incomplete),
            param("organization_follower_list", {}, marks=_NotAuthorized),
            ("organization_list_for_user", {}),
            ("organization_list", {}),
            ("organization_show", {"name": "tdci"}),
            param("package_autocomplete", {}, marks=_Incomplete),
            param("package_collaborator_list_for_user", {}, marks=_Incomplete),
            param("package_collaborator_list", {}, marks=_Incomplete),
            param("package_relationships_list", {}, marks=_Incomplete),
            ("package_search", {}),
            ("package_show", {"id": "8decd067-a38a-4834-a70c-40419b17fe9c"}),
            ("package_show", {"name": "2023-production-statistics"}),
            param("recently_changed_packages_activity_list", {}, marks=_NotImplemented),
            param("resource_search", {}, marks=_Incomplete),
            param("resource_show", {}, marks=_Incomplete),
            param("resource_view_list", {}, marks=_Incomplete),
            param("resource_view_show", {}, marks=_Incomplete),
            ("status_show", {}),
            ("tag_autocomplete", {}),
            ("tag_list", {}),
            ("tag_search", {}),
            ("tag_show", {"name": "transport"}),
            param("task_status_show", {}, marks=_Incomplete),
            param("term_translation_show", {}, marks=_Incomplete),
            param("user_autocomplete", {}, marks=_Incomplete),
            param("user_followee_count", {}, marks=_Incomplete),
            param("user_followee_list", {}, marks=_NotAuthorized),
            param("user_follower_count", {}, marks=_Incomplete),
            param("user_follower_list", {}, marks=_NotAuthorized),
            param("user_list", {}, marks=_NotAuthorized),
            param("user_show", {}, marks=_NotAuthorized),
            ("vocabulary_list", {}),
            param("vocabulary_show", {}, marks=_Incomplete),
        ],
    )
    def test_call_action_getattr(self, c, action, args) -> None:
        """Test calling actions accessed as class attributes."""
        method = getattr(c, action)

        result = method(args) if args else method()

        del result

    @pytest.mark.parametrize(
        "arg",
        (
            "tdci",
            "94861715-012e-4467-8353-ab6d1a9c32f9",
            {"name": "tdci"},
            {"id": "94861715-012e-4467-8353-ab6d1a9c32f9"},
            Organization(name="tdci"),
            Organization(id="94861715-012e-4467-8353-ab6d1a9c32f9"),
        ),
    )
    def test_show_action(self, c, arg) -> None:
        result = c.show_action(arg, _cls=Organization)
        del result


def test_get_class() -> None:
    assert None is get_class("foo")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import sys
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING

import transport_data  # noqa: F401

if TYPE_CHECKING:
    import sphinx.application

# -- Project information ---------------------------------------------------------------

project = "Tools for the Transport Data Commons (TDC)"
copyright = "2023–%Y, Transport Data Commons Initiative"
author = "Transport Data Commons Initiative"

# -- General configuration -------------------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_autorun",
]
nitpicky = True
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# A string of reStructuredText included at the beginning of every source file.
rst_prolog = r"""
.. role:: py(code)
   :language: python

.. role:: xml(code)
   :language: xml
"""


def setup(app: "sphinx.application.Sphinx"):
    from sphinx.ext.autosummary.generate import generate_autosummary_docs

    # NB this needs to be here because setup() is executed before the module-level
    #    variables are used
    app.config.templates_path.append("_template")
    generate_autosummary_docs(sources=["api.in"], output_dir="_api", app=app)


# -- Options for HTML output -----------------------------------------------------------

GH_URL = "https://github.com/transport-data/tools"

html_css_files = ["css/custom.css"]

html_theme = "furo"

html_logo = "_static/logo.png"

html_theme_options = dict(
    footer_icons=[
        {
            "name": "GitHub",
            "url": GH_URL,
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    top_of_page_buttons=["view", "edit"],
    source_repository=GH_URL,
    source_branch="main",
    source_directory="doc/",
)

html_title = "Tools for the Transport Data Commons (TDC)"

html_static_path = ["_static"]

# -- Options for sphinx.ext.autosummary ------------------------------------------------

autosummary_generate = True

# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/transport-data/tools/issues/%s", "#%s"),
    "portal-issue": ("https://github.com/transport-data/portal/issues/%s", "#%s"),
    "pull": ("https://github.com/transport-data/tools/pull/%s", "PR #%s"),
    "gh-user": ("https://github.com/%s", "@%s"),
}

# -- Options for sphinx.ext.intersphinx ------------------------------------------------

intersphinx_mapping = {
    "click": ("https://click.palletsprojects.com/en/8.1.x/", None),
    "docutils": ("https://sphinx-docutils.readthedocs.io/en/latest", None),
    "dsss": ("https://dsss.readthedocs.io/en/stable", None),
    "jinja2": ("https://jinja.palletsprojects.com/en/3.1.x", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "platformdirs": ("https://platformdirs.readthedocs.io/en/latest/", None),
    "pluggy": ("https://pluggy.readthedocs.io/en/stable", None),
    "pooch": ("https://www.fatiando.org/pooch/latest/", None),
    "py": ("https://docs.python.org/3/", None),
    "pytest": ("https://docs.pytest.org/en/stable/", None),
    "sdmx": ("https://sdmx1.readthedocs.io/en/stable/", None),
}

# -- Options for sphinx.ext.linkcode ---------------------------------------------------

base_path = {
    m: Path(import_module(m).__file__).parents[1]  # type: ignore [arg-type]
    for m in ("transport_data",)
}
base_url = "https://github.com/transport-data/tools/tree/main/"


def linkcode_resolve(domain, info):
    # TODO adjust `base_url` if not on `main` branch
    if domain != "py" or not info["module"]:
        return None

    # Module containing the object
    module = sys.modules[info["module"]]
    # Path of the module relative to the base path of its package
    rel = Path(module.__file__).relative_to(base_path[info["module"].split(".")[0]])

    # The object itself
    try:
        obj = getattr(module, info["fullname"])
    except AttributeError:
        return None
    try:
        # First line number
        firstlineno = obj.__code__.co_firstlineno
    except AttributeError:
        fragment = ""  # A variable, can't obtain the precise location
    else:
        fragment = f"#L{firstlineno}"  # Function, class, or other definition

    return f"{base_url}{rel}{fragment}"


# -- Options for sphinx.ext.napoleon ---------------------------------------------------

napoleon_preprocess_types = True

# -- Options for sphinx.ext.todo -------------------------------------------------------

todo_include_todos = True



import pytest
from requests.exceptions import SSLError

from transport_data.org.ckan import (
    DEV,
    PROD,
    STAGING,
    ckan_package_to_mdr,
    mdr_to_ckan_package,
)
from transport_data.testing import CKAN_UUID, CliRunner
from transport_data.util.ckan import Package


@pytest.fixture
def package(test_data_path) -> Package:
    """A :class:`.Package` from a test specimen."""
    return Package.from_file(test_data_path.joinpath("ckan", "package.json"))


@pytest.mark.parametrize(
    "instance, N_exp",
    [
        # (DEV, 451),
        (DEV, 5),
        # (PROD, 418),
        (PROD, 5),
        pytest.param(STAGING, None, marks=pytest.mark.xfail(raises=SSLError)),
    ],
)
def test_main(instance, N_exp: int) -> None:
    result = instance.package_list(max=5)

    assert N_exp <= len(result)

    p = instance.package_show(result[0])

    assert p is result[0]  # Same object instance returned

    assert "dataset" == p.type


def test_ckan_package_to_mdr(package: Package) -> None:
    """:func:`.ckan_package_to_mdr` works as expected."""
    from transport_data.org.metadata import _get

    # Function runs
    mdr = ckan_package_to_mdr(package)

    # Resulting MetadataReport has the expected number of attributes
    assert 48 == len(mdr.metadata)

    # Structured attributes are stored as their string representations
    assert repr(["cars", "private-cars", "truck", "bus"]) == _get(mdr, "modes")
    # Python types are stored as their representations
    assert "False" == _get(mdr, "is_archived")
    assert "True" == _get(mdr, "isopen")
    assert "None" == _get(mdr, "author")


def test_cli(tdc_cli: CliRunner) -> None:
    """:program:`tdc ckan foo` returns a non-zero exit code and displays error text."""
    result = tdc_cli.invoke(["ckan", "foo"])

    assert "Action name not known: foo" in result.output
    assert 1 == result.exit_code


@pytest.mark.ckan_dev
@pytest.mark.network
def test_create() -> None:  # pragma: no cover
    """A test package can be created on the DEV instance."""
    instance, owner_org = DEV, CKAN_UUID["DEV org test"]

    p = Package(
        # Minimal fields for creating a Package object
        name="transport-data-tools-test",  # TODO Create a hashed random name
        notes=".",
        owner_org=owner_org,
        tdc_category="public",
    )

    response = instance.package_create(**p.asdict())

    assert response


def test_mdr_to_ckan_package(package: Package) -> None:
    """:func:`.mdr_to_ckan_package` works as expected."""

    # Convert to a MetadataReport
    mdr = ckan_package_to_mdr(package)

    # Convert back to a Package instance
    p = mdr_to_ckan_package(mdr)

    if package != p:
        assert package.__dict__ == p.__dict__  # `pytest -vv` shows the rich comparison



"""European Commission Joint Research Centre (JRC) provider.

This module handles data from the JRC IDEES 2015 source.

"Handle" includes:

- Fetch
- Extract
- Parse the native spreadsheet layout.
- Convert to SDMX.
"""

import re
from collections import defaultdict
from functools import partial
from itertools import chain, count
from operator import add
from pathlib import Path
from typing import Any, Dict

import numpy as np
import pandas as pd
import sdmx.model.v21 as m

from transport_data import STORE
from transport_data.util.pluggy import hookimpl
from transport_data.util.pooch import Pooch
from transport_data.util.sdmx import anno_generated


@hookimpl
def get_agencies():
    """Return information about the agency providing the data set.

    See :func:`.org.get_agencyscheme`.
    """
    # Agency
    a = m.Agency(
        id="JRC",
        name="Joint Research Centre of the European Commission",
        description="""See https://joint-research-centre.ec.europa.eu/index_en.

Maintainers of the IDEES data set: https://data.jrc.ec.europa.eu/dataset/jrc-10110-10001""",  # noqa: E501
    )

    a.contact.append(
        m.Contact(name="Jacopo Tattini", email=["Jacopo.TATTINI@ec.europa.eu"])
    )

    return (a,)


@hookimpl
def provides():
    return ("ConceptScheme=TDCI:CS_JRC_MEASURE",)


#: Base URL for :data:`POOCH`.
BASE_URL = (
    "https://jeodpp.jrc.ec.europa.eu/ftp/jrc-opendata/JRC-IDEES/JRC-IDEES-2015_v1/"
)

#: List of geographical areas for which data are provided. These are ISO 3166-1 alpha-2
#: codes, except for "EU28".
GEO = [
    "AT",
    "BE",
    "BG",
    "CY",
    "CZ",
    "DE",
    "DK",
    "EE",
    "EL",
    "ES",
    "EU28",
    "FI",
    "FR",
    "HR",
    "HU",
    "IE",
    "IT",
    "LT",
    "LU",
    "LV",
    "MT",
    "NL",
    "PL",
    "PT",
    "RO",
    "SE",
    "SI",
    "SK",
    "UK",
]

# Fragments for file names in each archive
MEMBERS = [
    "EmissionBalance",
    "EnergyBalance",
    "Industry",
    "Macro",
    "MBunkers",
    "PowerGen",
    "Residential",
    "Tertiary",
    "Transport",
]


def expand(fname: str) -> str:
    """Callback to expand `fname` into a complete file name for :data:`POOCH`."""
    return f"JRC-IDEES-2015_All_xlsx_{fname}.zip"


#: Pooch to fetch data from the JRC website at :data:`BASE_URL`.
POOCH = Pooch(
    module=__name__,
    base_url=BASE_URL,
    # TODO add copyright.txt, readme.txt, and "…Methodological Note.pdf"
    registry={expand(geo): None for geo in GEO},
    expand=expand,
    processor="unzip",
)


def fetch(*geo, dry_run: bool = False):
    if dry_run:
        for g in geo:
            print(f"Valid url for GEO={g}: {POOCH.is_available(g)}")
        return

    return list(chain(*[POOCH.fetch(g) for g in geo]))


def path_for(geo=None, member=None):
    """Return a filename and local cache path for the data file for `geo`."""
    if member:
        assert member in MEMBERS
        return POOCH.path.joinpath(f"JRC-IDEES-2015_{member}_{geo}.xlsx")
    else:
        return POOCH.path.joinpath(expand(geo))


def iter_blocks(path: Path, geo: str):
    ef = pd.ExcelFile(path)

    skip_sheets = ["cover", "index"]
    parse_args: Dict[Any, Any] = dict(usecols="A:Q", header=None)
    columns = pd.Index([])
    for sheet_name in map(str, filter(lambda n: n not in skip_sheets, ef.sheet_names)):
        common = dict(GEO=geo)

        # Extract a hint for MODE from the sheet name
        if match := re.match("Tr(Avia|Navi|Rail|Road)_...", sheet_name):
            assert match
            value = match.group(1).upper()
            common.update(MODE={"AVIA": "AIR", "NAVI": "WATER"}.get(value, value))

        df = ef.parse(sheet_name, **parse_args)

        # Identify blank rows that separate blocks of data
        breaks = df.index[df.isna().all(axis=1)].to_list()

        # Iterate over blocks
        for start, end in zip(map(partial(add, 1), [-1] + breaks), breaks):
            data = df.loc[start : end - 1, :]
            if start == 0:
                assert 1 == data.shape[0]
                columns = pd.Index(["INFO"] + data.iloc[0, 1:].astype(int).to_list())
                continue
            elif data.iloc[0, 0] == "Indicators":
                assert 1 == data.shape[0] and data.iloc[0, 1:].isna().all()
                continue

            yield data.pipe(_set_axis, columns).assign(**common).reset_index(drop=True)


def _set_axis(df: pd.DataFrame, columns: pd.Index) -> pd.DataFrame:
    melt = partial(pd.DataFrame.melt, id_vars="INFO", value_name="OBS_VALUE")

    if df.iloc[0, 0] is np.nan:
        # Handle a different structure used in TrRoad_tech
        # First row contains the single TIME_PERIOD label applying to all data
        tp = df.iloc[0, :].dropna().item()
        # Construct a different set of columns from the *second* row of `df`, and only
        # carry through the third row et seq. that contain actual data
        return (
            df.iloc[2:, :]
            .set_axis(pd.Index(["INFO"] + df.iloc[1, 1:].to_list()), axis=1)
            .pipe(melt, var_name="YEAR_REG")
            .assign(TIME_PERIOD=tp)
        )
    else:
        return df.set_axis(columns, axis=1).pipe(melt, var_name="TIME_PERIOD")


def _match_extract(df: pd.DataFrame, expr) -> pd.DataFrame:
    cols = list(re.compile(expr).groupindex.keys())

    matches = df["INFO"].str.fullmatch(expr)
    result = pd.concat([df, df["INFO"].str.extract(expr)[cols].ffill()], axis=1)
    result.loc[matches, "INFO"] = "ALL"
    return result


def _fill_unit_measure(df: pd.DataFrame, measure: str) -> pd.DataFrame:
    """Fill in the UNIT_MEASURE column of `df`."""
    # Identify the MODE, if any
    mode = df.get("MODE", default=pd.Series([None])).unique().item()  # type: ignore
    return df.assign(
        UNIT_MEASURE=lambda df: df[["U0", "U1"]]
        .fillna(UNIT_MEASURE.get((mode, measure), ""))
        .bfill(axis=1)["U0"]
    ).drop(["U0", "U1"], axis=1)


#: Mapping from (MODE, MEASURE) to UNIT_MEASURE attribute value, where these are not
#: specified in the data.
UNIT_MEASURE = {
    # Both TrRoad_act and TrRoad_tech
    ("ROAD", "New vehicle-registrations"): "vehicle",
    # TrRoad_tech
    ("ROAD", "Age structure in 2015"): "vehicle",
    # TrRail_act
    ("RAIL", "Occupancy / utilisation"): "percent",
    # TrAvia_act
    ("AIR", "Number of flights"): "1",
    ("AIR", "Stock of aircrafts - total"): "vehicle",
    ("AIR", "Stock of aircrafts - in use"): "vehicle",
    ("AIR", "New aircrafts"): "vehicle",
    ("AIR", "Flights per year by airplance"): "1",
    # TrAvia_png
    ("AIR", "Number of seats available"): "1",
    ("AIR", "Seats available per flight"): "1",
}

UNPACK: Dict[str, Dict[str, str]] = {
    "ALL": dict(),
    "Road transport": dict(MODE="ROAD", VEHICLE_TYPE="_T"),
    "Powered 2-wheelers": dict(MODE="ROAD", VEHICLE_TYPE="2W"),
    "Powered 2-wheelers (Gasoline)": dict(
        MODE="ROAD", VEHICLE_TYPE="2W", FUEL="Gasoline"
    ),
    "of which biofuels": dict(FUEL="Biofuels"),
    "Passenger cars": dict(MODE="ROAD", VEHICLE_TYPE="LDV"),
    "Gasoline engine": dict(MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="GAS"),
    "Diesel oil engine": dict(MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="DIES"),
    "LPG engine": dict(MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="LPG"),
    "Natural gas engine": dict(MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="NG"),
    "of which biogas": dict(FUEL="Biogas"),
    "Plug-in hybrid electric": dict(MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="PHEV"),
    "Plug-in hybrid electric (Gasoline and electricity)": dict(
        MODE="ROAD",
        VEHICLE_TYPE="LDV",
        POWERTRAIN="PHEV",
        FUEL="ALL",
    ),
    "of which electricity": dict(FUEL="ELE"),
    "Battery electric vehicles": dict(
        MODE="ROAD", VEHICLE_TYPE="LDV", POWERTRAIN="BEV"
    ),
    "Motor coaches, buses and trolley buses": dict(MODE="ROAD", VEHICLE_TYPE="BUS"),
    "Rail, metro and tram": dict(MODE="RAIL", VEHICLE_TYPE="ALL"),
    "Metro and tram, urban light rail": dict(MODE="RAIL", VEHICLE_TYPE="L"),
    "Conventional passenger trains": dict(MODE="RAIL", VEHICLE_TYPE="H"),
    "High speed passenger trains": dict(MODE="RAIL", VEHICLE_TYPE="HS"),
    "Aviation": dict(MODE="AIR", SEGMENT="ALL"),
    "Domestic": dict(MODE="AIR", SEGMENT="DOM"),
    "International - Intra-EU": dict(MODE="AIR", SEGMENT="IN_EU"),
    "International - Extra-EU": dict(MODE="AIR", SEGMENT="EX_EU"),
    # Freight
    "Light duty vehicles": dict(MODE="ROAD", VEHICLE_TYPE="LDV"),
    "Heavy duty vehicles": dict(MODE="ROAD", VEHICLE_TYPE="HDV"),
    "Heavy duty vehicles (Diesel oil incl. biofuels)": dict(
        MODE="ROAD",
        VEHICLE_TYPE="HDV",
        FUEL="Diesel oil incl. biofuels",
    ),
    "Rail transport": dict(MODE="RAIL", VEHICLE_TYPE="_T"),
    "Domestic and International - Intra-EU": dict(MODE="RAIL", SEGMENT="DOM_IN_EU"),
    "Coastal shipping and inland waterways": dict(MODE="WATER", SEGMENT="ALL"),
    "Domestic coastal shipping": dict(MODE="WATER", SEGMENT="DOM"),
    "Inland waterways": dict(MODE="WATER", SEGMENT="IWW"),
    # Sheet TrRoad_ene, table "Total energy consumption (ktoe)"
    # "Domestic": dict(MODE="ROAD", VEHICLE_TYPE="TRUCK", SEGMENT="DOM"),
    "International": dict(SEGMENT="INTL"),
    # Sheet TrRoad_ene
    "by fuel": dict(FUEL="ALL"),
    "by fuel (EUROSTAT DATA)": dict(FUEL="ALL"),
    "Liquids": dict(FUEL="Liquids"),
    "Liquids (Petroleum products)": dict(FUEL="Liquids (without biofuels)"),
    "Liquified petroleum gas (LPG)": dict(FUEL="LPG"),
    "LPG": dict(FUEL="LPG"),
    "Gasoline (without biofuels)": dict(FUEL="Gasoline (without biofuels)"),
    "Gasoline (incl. biofuels)": dict(FUEL="Gasoline (incl. biofuels)"),
    "Gas/Diesel oil (without biofuels)": dict(FUEL="Gas/Diesel oil (without biofuels)"),
    "Diesel": dict(FUEL="Diesel"),
    "Diesel oil": dict(FUEL="Diesel"),
    "Diesel oil (incl. biofuels)": dict(FUEL="Diesel oil (incl. biofuels)"),
    "Kerosene": dict(FUEL="Kerosene"),
    "Residual fuel oil": dict(FUEL="Residual fuel oil"),
    "Other petroleum products": dict(FUEL="Other petroleum products"),
    "Natural gas": dict(FUEL="NG"),
    "Natural gas (incl. biogas)": dict(FUEL="Natural gas (incl. biogas)"),
    "Renewable energies and wastes": dict(FUEL="Renewable energies and wastes"),
    "Biogas": dict(FUEL="Biogas"),
    "Biogasoline": dict(FUEL="Biogasoline"),
    "Biodiesel": dict(FUEL="Biodiesel"),
    "Biomass and wastes": dict(FUEL="Biomass and wastes"),
    "Other biofuels": dict(FUEL="Other biofuels"),
    "Electricity": dict(FUEL="ELE"),
    "Electric": dict(FUEL="ELE"),
    "Solids": dict(FUEL="Solids"),
}


def _unpack_info(df: pd.DataFrame) -> pd.DataFrame:
    """Unpack values from the INFO column."""
    info = df["INFO"].drop_duplicates()
    try:
        unpacked = pd.concat([info, info.apply(lambda v: pd.Series(UNPACK[v]))], axis=1)
    except KeyError as e:
        print(f"Failed to unpack INFO for {e.args[0]!r}:")
        print(info.to_string())
        assert False

    def _merge_mode(df: pd.DataFrame) -> pd.DataFrame:
        cols = ["MODE_x", "MODE_y"]
        try:
            return df.assign(MODE=df[cols].ffill(axis=1)[cols[-1]]).drop(cols, axis=1)
        except KeyError:
            return df

    return (
        df.merge(unpacked, on="INFO")
        .drop("INFO", axis=1)
        .pipe(_merge_mode)
        .fillna("_X")
    )


def read(geo=None):
    """Read data from a single file.

    Notes on the format:

    - The hierarchy of some codes in column A is expressed by Excel formatting
      (indentation, with greater indentation indicating greater depth). This is not
      easily extracted with pandas.
    """
    path = path_for(geo, "Transport")

    # Metadata across all blocks
    CS_MEASURE = m.ConceptScheme(id="MEASURE")
    _measure_id = dict()
    # Data per measure
    data = defaultdict(list)

    print(f"Read data for {geo = }")

    # Iterate over blocks of data
    i = count(start=1)
    for block in iter_blocks(path, geo):
        # Identify the measure (INFO column on first row of the block)
        measure_unit = block.loc[0, "INFO"]

        # Unpack a string that combines MEASURE and, possibly, UNIT_MEASURE
        match = re.fullmatch(r"(?P<measure>[^\(]+) \((?P<unit>.*)\)\*?", measure_unit)
        try:
            measure, unit = match.group("measure", "unit")
        except AttributeError:
            measure, unit = measure_unit, None

        # The measure concept
        try:
            mc_id = _measure_id[measure]
        except KeyError:
            mc_id = f"{next(i):02d}"
            CS_MEASURE.append(m.Concept(id=mc_id, name=measure))
            _measure_id[measure] = mc_id

        # - Assign `unit` from above as U0, one candidate from UNIT_MEASURE.
        # - Replace the embedded measure/unit expression with "ALL transport".
        # - Unpack and fill forward SERVICE values from "Passenger transport" or
        #   "Freight transport" entries in the INFO column. Also unpack U1.
        # - Choose the right
        # - If UNIT_MEASURE is still missing, fill from explicit values above.
        # - Drop rows with missing values in "OBS_VALUE"
        block = (
            block.assign(U0=unit)
            .replace({"INFO": {measure_unit: "ALL transport"}})
            .pipe(
                _match_extract,
                r"(?P<SERVICE>ALL|Freight|Passenger) transport( +\((?P<U1>.*)\))?",
            )
            .pipe(_fill_unit_measure, measure)
            .dropna(subset="OBS_VALUE", ignore_index=True)
            .pipe(_unpack_info)
        )

        data[mc_id].append(block)

    print(f"{sum(map(len, data.values()))} data sets for {len(CS_MEASURE)} measures")

    # Join multiple data frames
    return {
        CS_MEASURE[mc_id]: pd.concat(dfs).fillna("_X") for mc_id, dfs in data.items()
    }


def convert(geo):
    data = read(geo)

    # TODO create SDMX DSD, Dataflow, and DataSets from the resulting pd.DataFrame

    # Code lists
    CL = dict()

    # Retrieve the parent concept scheme
    CS_MEASURE = list(data.keys())[0].parent

    # Iterate over measure_concepts
    for measure_concept, df in data.items():
        # Update the structure information
        for concept, values in df.items():
            if concept in ("GEO", "OBS_VALUE", "TIME_PERIOD"):
                continue

            cl = CL.setdefault(concept, m.Codelist(id=concept))

            for value in values.unique():
                try:
                    cl.append(m.Code(id=value, name=value))
                except ValueError:
                    pass  # Already exists

        # Prepare an empty data set, associated structures, and a helper function
        dims = []
        for c in df.columns:
            if c in ("OBS_VALUE",):
                continue
            dims.append(c)
        ds, _make_obs = prepare(measure_concept, dims)

        # Convert rows of `data` into SDMX Observation objects
        ds.obs.extend(_make_obs(row) for _, row in df.iterrows())
        assert len(ds) == len(df)

        # Write the data set, DSD, and DFD to file
        for obj in (ds, ds.described_by, ds.structured_by):
            obj.annotations.append(
                m.Annotation(
                    id="tdc-comment", text=f"Primary measure is {measure_concept!r}"
                )
            )
            STORE.set(obj)

    # Write code lists, measure concept scheme to file
    a = get_agencies()[0]
    for obj in chain(CL.values(), [CS_MEASURE]):
        obj.maintainer = a
        obj.version = "0.1.0"
        STORE.set(obj)

    raise NotImplementedError("Merging data for multiple GEO")


def prepare(measure_concept, dims):
    # TODO merge with the similar function in .ato.__init__

    measure_id = measure_concept.id
    c = measure_concept
    aa = measure_concept

    # Data structure definition with an ID matching the measure
    # NB here we set JRC as the maintainer. Precisely, JRC establishes the data
    #    structure, but TDCI is maintaining the SDMX representation of it.
    dsd = m.DataStructureDefinition(
        id=measure_id, maintainer=get_agencies()[0], version="0.0.0"
    )
    anno_generated(dsd)

    dfd = m.DataflowDefinition(
        id=measure_id, maintainer=get_agencies()[0], version="0.0.0", structure=dsd
    )

    pm = m.PrimaryMeasure(id="OBS_VALUE", concept_identity=c)
    dsd.measures.append(pm)

    # Dimensions
    dsd.dimensions.extend(m.Dimension(id=d) for d in dims)

    # Convert annotations to DataAttributes. "NoSpecifiedRelationship" means that the
    # attribute is attached to an entire data set (not a series, individual obs, etc.).
    da = {}  # Store references for use below
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        da[a.id] = m.DataAttribute(id=a.id, related_to=m.NoSpecifiedRelationship())
        dsd.attributes.append(da[a.id])

    _PMR = m.PrimaryMeasureRelationship  # Shorthand

    # Convert remark column labels to DataAttributes. "PrimaryMeasureRelationship" means
    # that the attribute is attached to individual observations.
    for name in aa.eval_annotation("remark-cols") or []:
        dsd.attributes.append(m.DataAttribute(id=name, related_to=_PMR()))

    # Empty data set structured by this DSD

    ds = m.StructureSpecificDataSet(described_by=dfd, structured_by=dsd)
    try:
        ds.annotations.append(aa.get_annotation(id="remark-cols"))
    except KeyError:
        pass
    anno_generated(ds)

    # Convert temporary annotation values to attributes attached to the data set
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        ds.attrib[a.id] = m.AttributeValue(value=str(a.text), value_for=da[a.id])

    def _make_obs(row):
        """Helper function for making :class:`sdmx.model.Observation` objects."""
        key = dsd.make_key(m.Key, row[[d.id for d in dsd.dimensions]])

        # Attributes
        attrs = {}
        for a in filter(lambda a: a.related_to is _PMR, dsd.attributes):
            # Only store an AttributeValue if there is some text
            value = row[a.id]
            if not pd.isna(value):
                attrs[a.id] = m.AttributeValue(value_for=a, value=value)

        return m.Observation(
            dimension=key, attached_attribute=attrs, value_for=pm, value=row[pm.id]
        )

    return ds, _make_obs



import json
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path

import click
from platformdirs import user_cache_path, user_config_path, user_data_path


@dataclass
class Config:
    """Common configuration."""

    #: Path for local cache.
    cache_path: Path = field(
        default_factory=lambda: user_cache_path("transport-data", ensure_exists=True)
    )

    #: Path to the configuration file read, if any.
    config_path: Path | None = None

    #: Path for local data. See :ref:`store-layout` for details.
    data_path: Path = field(
        default_factory=lambda: user_data_path("transport-data", ensure_exists=True)
    )

    #: Git remote URL for the TDC SDMX registry.
    #:
    #: Users with SSH keys registered on github.com may wish to use instead
    #: ``git@github.com:transport-data/registry.git``.
    registry_remote_url: str = "https://github.com/transport-data/registry.git"

    #: Mapping from maintainer IDs to either "local" (stored in a :class:`.LocalStore`)
    #: or "registry" (stored in the :class:`.Registry`).
    store_map: dict = field(
        default_factory=lambda: {"TDCI": "registry", "TEST": "local"}
    )

    @staticmethod
    def _config_path():
        """Candidate path for the configuration file."""
        return user_config_path("transport-data").joinpath("config.json")

    @classmethod
    def read(cls):
        """Read configuration from file, returning a new instance."""
        cp = cls._config_path()
        if cp.exists():
            with open(cp) as f:
                data = json.load(f)
            data.update(config_path=cp)
        else:
            # No file exists → defaults for all settings
            data = dict()

        return cls(**data)

    def write(self):
        """Write configuration to file."""
        cp = self._config_path()
        if not cp.parent.exists():
            print(f"Create {cp.parent}")
            cp.parent.mkdir(parents=True, exist_ok=True)

        # Convert dataclass instance to dict; omit the path to the file itself
        data = asdict(self)
        data.pop("config_path")
        for name in "cache_path", "data_path":
            data[name] = str(data[name])

        # Write to config.json
        cp.write_text(json.dumps(data, indent=2))
        print(f"Wrote {cp}")


@click.group("config")
def main():
    """Manipulate configuration."""
    pass


@main.command("set")
@click.argument("key")
@click.argument("value")
def set_cmd(key, value):
    """Set config KEY to VALUE."""
    from transport_data import CONFIG

    # Validate `key`
    field_names = [f.name for f in fields(CONFIG)]
    if key not in field_names:
        raise click.ClickException(
            f"{key!r} not among configuration keys {field_names}"
        )

    # Set the value
    setattr(CONFIG, key, value)

    # Write the configuration file
    CONFIG.write()


@main.command()
def show():
    """Display config values."""
    from transport_data import CONFIG

    print(*[f"{k}: {v}" for k, v in asdict(CONFIG).items()], sep="\n")



"""Utilities for :mod:`sdmx`."""

import io
from dataclasses import fields
from datetime import datetime
from importlib.metadata import version
from inspect import getsource
from io import StringIO
from textwrap import dedent
from token import COMMENT, NAME
from tokenize import generate_tokens
from typing import TYPE_CHECKING, cast

import pandas as pd

if TYPE_CHECKING:
    import pathlib
    from typing import TypedDict

    import sdmx.model.common
    import sdmx.model.v21
    import sdmx.model.v30

    # Types for dicts holding keyword arguments to SDMX classes
    # TODO Move upstream
    class IAKeywords(TypedDict, total=False):
        id: str

    class VAKeywords(IAKeywords):
        version: str

    class MAKeywords(VAKeywords):
        maintainer: sdmx.model.common.Agency | None


class CSVAdapter(io.RawIOBase):
    """Adapt CSV content from `path` into SDMX-CSV.

    The `SDMX-CSV <https://github.com/sdmx-twg/sdmx-csv/blob/master/data-message/docs/sdmx-csv-field-guide.md>`_
    format has precise requirements for identifier and other columns. In particular, the
    "STRUCTURE", "STRUCTURE_ID", and "ACTION" columns are mandatory.

    .. code-block::

       STRUCTURE,STRUCTURE_ID,ACTION,DIM_1,DIM_2,DIM_3,OBS_VALUE
       dataflow,ESTAT:NA_MAIN(1.6.0),I,A,B,2014-01,12.4

    This class produces standard SDMX-CSV by adapting a simplified format such as:

    .. code-block::

       DIM_1,DIM_2,DIM_3,OBS_VALUE
       A,B,2014-01,12.4

    In particular:

    - The header line of `path` is prefixed with "STRUCTURE", "STRUCTURE_ID", and/or
      "ACTION" if the respective parameters are given.
    - Every record (line) of `path` is prefixed with the *values* of the respective
      parameters.

    So, for instance, to transform the above example into the first example, use a
    CSVAdapter with:

    .. code-block:: python

       a = CSVAdapter(
           path,
           structure="dataflow",
           structure_id="ESTAT:NA_MAIN(1.6.0)",
           action="I",
       )
       data = a.read()

    Parameters
    ----------
    structure :
        Value for the "STRUCTURE" SDMX-CSV field, to be inserted into every record.
    structure_id :
        Value for the "STRUCTURE_ID" field, to be inserted into every record.
    action :
        Value for the "ACTION" field, to be inserted into every record.
    """

    def __init__(
        self,
        path: "pathlib.Path",
        structure: str | None = None,
        structure_id: str | None = None,
        action: str | None = None,
    ) -> None:
        self._path = path

        # Determine fields to prefix to header line and records
        header, line = [], []
        for column, value in (
            (b"STRUCTURE", structure),
            (b"STRUCTURE_ID", structure_id),
            (b"ACTION", action),
        ):
            if value is not None:
                header.append(column)
                line.append(value.encode())

        # Construct single `bytes`` for header and line prefixes
        self._header_prefix = b",".join((header + [b""]) if len(header) else [])
        self._line_prefix = b",".join((line + [b""]) if len(line) else [])

    def readall(self, size: int = -1, /) -> bytes:
        """Read and adapt CSV to SDMX-CSV.

        .. todo:: This currently reads and adapts the entire file at once; this may not
           be performant for very large files. Improve to handle lines individually or
           in batches.
        """
        lines = []
        with open(self._path, "rb") as f:
            lines = [
                (self._header_prefix if i == 0 else self._line_prefix) + line
                for i, line in enumerate(f)
            ]

        return b"".join(lines)


def anno_generated(obj: "sdmx.model.common.AnnotableArtefact") -> None:
    """Annotate the `obj` with information about how it was generated."""
    from sdmx.model import v21 as m

    try:
        # Retrieve existing annotation
        anno = obj.get_annotation(id="tdc-generated")
    except KeyError:
        # Create a new annotation
        anno = m.Annotation(id="tdc-generated")
        obj.annotations.append(anno)

    anno.text = (
        f"{datetime.now().isoformat()} by transport_data v{version('transport_data')}"
    )


def make_obs(
    row: "pd.Series", dsd: "sdmx.model.v21.DataStructureDefinition"
) -> "sdmx.model.v21.Observation":
    """Helper function for making :class:`sdmx.model.Observation` objects."""
    from sdmx.model import v21 as m

    key = dsd.make_key(m.Key, row[[d.id for d in dsd.dimensions]].to_dict())

    # Attributes
    attrs = {}
    for a in filter(
        lambda a: isinstance(a.related_to, m.PrimaryMeasureRelationship), dsd.attributes
    ):
        # Only store an AttributeValue if there is some text
        value = row[a.id]
        if not pd.isna(value):
            attrs[a.id] = m.AttributeValue(value_for=a, value=value)

    pm = dsd.measures[0]
    return m.Observation(
        dimension=key, attached_attribute=attrs, value_for=pm, value=row[pm.id]
    )


def read_csv(
    path: "pathlib.Path",
    structure: "sdmx.model.v30.Dataflow | sdmx.model.v30.DataStructureDefinition",
    adapt: dict | None = None,
) -> "sdmx.message.DataMessage":
    """Read or adapt SDMX-CSV from `path`.

    Parameters
    ----------
    path :
        A file in SDMX-CSV or CSV format.
    structure :
        Data flow or data structure describing the contents of `path`.
    adapt :
        Keyword arguments to :class:`CSVAdapter`. If given, the contents of `path` are
        adapted from a ‘simplified’ or ‘reduced’ CSV format to SDMX-CSV on-the-fly. See
        the class documentation for details.
    """
    import sdmx

    if adapt:
        source: "pathlib.Path" | "CSVAdapter" = CSVAdapter(path, **adapt)
    else:
        source = path

    return cast(
        "sdmx.message.DataMessage",
        sdmx.read_sdmx(source, format="csv", structure=structure),
    )


def fields_to_mda(
    cls: type,
    rs: "sdmx.model.v21.ReportStructure",
    cs: "sdmx.model.common.ConceptScheme | None" = None,
) -> None:
    """Populate `rs` with MetadataAttributes corresponding to dataclass fields of `cls`.

    Examples
    --------
    >>> @dataclass
    ... class MDSExample:
    ...     #: Foo
    ...     #:
    ...     #: Description of Foo.
    ...     foo: str
    ...
    ...     bar: int
    ...
    ... fields_to_mda(MDSExample)

    In this example, two metadata attributes will be added to `rs`:

    1. With id="foo" and an annotation with id="data-type" and text="<class 'str'>".
       The concept identity for the metadata attribute will also have id="foo",
       name="Foo", and description="Description of Foo."
    2. With id="bar" and an annotation with id="data-type" and text="<class 'int'>".
    """
    from sdmx.model import common, v21

    # Assemble info about the dataclass fields of `cls`
    field_info = {f.name: (f, "") for f in fields(cls)}

    # Tokenize the source code of `cls` and update `field_info` with the Sphinx-style
    # comments that precede each of the fields
    #
    # Thanks to https://davidism.com/attribute-docstrings/ and
    # https://stackoverflow.com/a/7457047
    comments = []  # Accumulate comment tokens
    for tok in generate_tokens(StringIO(dedent(getsource(cls))).readline):
        if tok.type == COMMENT:
            comments.append(tok.string.lstrip("#: "))  # Store
        elif tok.type == NAME and tok.string in field_info and len(comments):
            # Reached the definition of a field, and there are accumulated comments
            field_info[tok.string] = (field_info[tok.string][0], "\n".join(comments))
            comments = []  # Reset

    cs = cs or common.ConceptScheme()

    for id_, (f, docstring) in field_info.items():
        # Split the docstring, if any, to a name and optional description
        name, _, desc = docstring.partition("\n\n")

        # Construct the ConceptIdentity and add to `cs`
        ci = cs.setdefault(id=id_, name=name or None, description=desc or None)
        # Construct the data type annotation
        type_anno = v21.Annotation(id="data-type", text={"zxx": repr(f.type)})
        # Add the metadata attribute to the report structure
        rs.getdefault(id=id_, concept_identity=ci, annotations=[type_anno])



import os

import pytest


@pytest.fixture(scope="session")
def test_config(tmp_path_factory):
    name = "XDG_CONFIG_HOME"
    pre = os.environ.get(name)
    os.environ[name] = str(tmp_path_factory.mktemp("dot-config"))

    yield

    if pre is None:
        os.environ.pop(name)
    else:
        os.environ[name] = pre


usefixtures = test_config



from typing import TYPE_CHECKING

import click

from transport_data.org.ckan import instance_option

if TYPE_CHECKING:
    from transport_data.util.ckan import Client


@click.group("test")
def main():
    """Utilities for testing and development."""


@main.command(params=[instance_option])
def wipe(instance: "Client"):  # pragma: no cover
    """Remove all packages in the CKAN test organization.

    This requires a configured API token with permissions to delete packages.
    """
    from transport_data.testing import CKAN_UUID

    client = instance

    result = client.organization_show(
        id=CKAN_UUID[f"{client.id.upper()} org test"], include_datasets=True
    )
    packages = result.packages
    if not packages:
        print(f"{result!r} contains no packages; nothing to do")
        return

    print(f"Will delete {len(packages)} packages:")
    for i, p in enumerate(packages, start=1):
        print(f"{i:3d}. {p!r}")

    user_input = input("\nType 'yes' to confirm: ")

    if user_input != "yes":
        return

    deleted = []
    for p in packages:
        client.package_delete(id=p.id)
        deleted.append(p)

    print(f"\nDeleted {len(packages)} packages")



"""Command-line interface."""

from importlib import import_module
from pathlib import Path

import click

from transport_data import CONFIG  # noqa: F401


@click.group("tdc")
def main():
    """Transport Data Commons tools."""


#: List of (sub)modules that define CLI (sub)commands. Each should contain a
#: @click.command() named "main".
MODULES_WITH_CLI = [
    "ato.cli",
    "config",
    "cli.interactive",
    "estat",
    "iamc.cli",
    "iso.cli",
    "itdp.cli",
    "jrc.cli",
    "oica.cli",
    "org.cli",
    "org.ckan",
    "proto.cli",
    "store",
    "testing.cli",
]


# Add commands from each module that defines them
for name in MODULES_WITH_CLI:
    module = import_module(f"transport_data.{name}")
    main.add_command(getattr(module, "main"))


@main.command()
@click.argument("structure_urn", metavar="URN")
@click.argument(
    "path", metavar="FILE", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.option("--sheets", help="Sheet(s) in .xlsx FILE to check.")
@click.option("-v", "--verbose", count=True, help="Increase verbosity.")
@click.option("--structure", help="Value for STRUCTURE field.")
@click.option("--structure-id", "structure_id", help="Value for STRUCTURE_ID field.")
@click.option("--action", default="I", help="Value for ACTION field.")
def check(structure_urn: str, path: Path, sheets, verbose, **options):  # noqa: C901
    """Check that FILE can be read as SDMX-CSV.

    URN is the shortened SDMX URN of a data flow or data structure definition that
    describes the data in FILE, for example "Dataflow=PROVIDER:EXAMPLE(1.2.3)" (the
    version is not required). This artefact must already be present in the local store.

    FILE may have a ".csv" or ".xlsx" suffix. In the latter case, it is converted to a
    temporary set of CSV files. If --sheets are given, only these worksheets are
    converted and checked.

    If not given, --structure and --structure-id are inferred from URN.
    """
    from traceback import format_exception

    import sdmx
    import sdmx.urn
    from sdmx.model import common

    from transport_data import STORE
    from transport_data.util.sdmx import read_csv

    # Pieces of any error message
    message = []

    # Handle `structure_urn`: retrieve a data structure that describes the data
    try:
        structure = STORE.get(structure_urn)
    except Exception:
        message.append(f"Structure {structure_urn!r} could not be loaded")
        structure = structure_cls = structure_id = None
    else:
        structure_cls = type(structure).__name__.lower().replace("definition", "")
        structure_id = sdmx.urn.shorten(structure.urn).split("=")[-1]

    if isinstance(structure, common.BaseDataflow):
        # Also retrieve the data structure definition
        STORE.resolve(structure, "structure")
        assert len(structure.structure.dimensions)

    # Construct keyword arguments for CSVAdapter
    # TODO Check if this works for full SDMX-CSV
    adapt = {
        "structure": options.pop("structure") or structure_cls,
        "structure_id": options.pop("structure_id") or structure_id,
        "action": options.pop("action"),
    }

    # Handle `path`; construct a sequence of (label, path) of CSV files to be processed
    label_path = []

    if path.suffix == ".csv":
        label_path.append((f"File: {path}", path))
    elif path.suffix == ".xlsx":
        # Explode an Excel file into one or more CSV files in a temporary directory
        import pandas as pd
        from platformdirs import user_cache_path

        # Create a cache directory
        cache_dir = user_cache_path("transport-data").joinpath("check")
        cache_dir.mkdir(parents=True, exist_ok=True)

        # Explode Excel file into one CSV file per sheet
        ef = pd.ExcelFile(path)
        _sheets = set(sheets.split(",")) if sheets else set(ef.sheet_names)
        for sheet_name in filter(_sheets.__contains__, ef.sheet_names):
            # Construct a temporary path
            label_path.append(
                (
                    f"File: {path}\nSheet: {sheet_name}",
                    cache_dir.joinpath(f"{path.stem}_xlsx_{sheet_name}.csv"),
                )
            )
            # Read the sheet from the ExcelFile and write to a CSV file
            pd.read_excel(ef, sheet_name).to_csv(label_path[-1][1], index=False)
        ef.close()
    else:
        raise click.UsageError(f"Unsupported file extension: {path.suffix!r}")

    # Process `label_path`
    for label, p in label_path:
        print(f"\n{label}")

        # Read the file into an SDMX data message
        try:
            dm = read_csv(p, structure, adapt)
        except Exception as e:
            message.append(f"read failed with\n{type(e).__name__}: {' '.join(e.args)}")

            if len(e.args) and "line 1" in e.args[0]:
                message.append(
                    "Hint: try giving --structure= or --structure-id argument(s) to "
                    "adapt to SDMX-CSV."
                )
            elif structure is None:
                pass
            else:  # pragma: no cover
                message.append("\n".join(format_exception(e)))

            print("")
            raise click.ClickException("\n\n".join(message))

        # Show the contents of the data message
        dfd_urn = sdmx.urn.shorten(sdmx.urn.make(dm.dataflow))
        print(f"\n{len(dm.data)} data set(s) in: {dfd_urn!s}")

        # Show information about each data set
        for i, ds in enumerate(dm.data):
            print(f"\nData set {i}: action={ds.action}")

            # Show the data set contents or summary, according to verbosity
            if verbose == 0:
                print(f"{len(ds)} observations")
            elif verbose == 1:
                print(sdmx.to_pandas(ds))
            else:
                print(sdmx.to_pandas(ds).to_string())



"""CLI for :mod:`.iamc`.

.. runblock:: console

   $ tdc iamc --help

"""

from pathlib import Path

import click


@click.command(name="iamc")
def main():
    """Demonstrate IAMC structure generation.

    This command uses a test snippet of data from the IPCC AR6 database.
    """
    import pandas as pd

    from transport_data import registry
    from transport_data.iamc import structures_for_data

    # Load test data
    test_data_path = Path(__file__).parents[1].joinpath("data", "tests")
    df = pd.read_csv(test_data_path.joinpath("iamc.csv"))

    # Function runs, returns a SDMX StructureMessage containing multiple structure
    # objects
    sm = structures_for_data(df, base_id="TEST")

    # Write each of the structure objects received to a separate file
    registry.write(sm, annotate=True, force=True)



"""Non-standard TDC Excel file format for collecting metadata."""

import logging
import re
from typing import TYPE_CHECKING, List, Tuple, cast

from sdmx.model import common, v21

if TYPE_CHECKING:
    import pathlib

    from openpyxl import Workbook
    from openpyxl.cell.cell import Cell
    from openpyxl.worksheet.worksheet import Worksheet

    from transport_data.util.sdmx import MAKeywords

log = logging.getLogger(__name__)


#: README text for the TDC metadata file format.
README_TEXT = """This file is an unofficial, prototype TDC format for metadata.
loosely imitates the Eurostat format. These files contain metadata (information
*about* data) based on the SDMX information model, but their layout (sheet
names, columns, etc.) is not specified by the SDMX standard, hence ‘unofficial’.

This file has the following sheets.

README
======

This sheet.

Attributes
==========

- One row per metadata attribute (or 'field').
- Columns for the name; description; and ID (short and machine-readable) of each
  attribute. See these descriptions to learn what to write for each attribute.

One or more additional sheets
=============================

- The name (or title) of each sheet corresponds to the identity (ID) of the data
  flow that is described by the metadata in that sheet.
- In Column A, the name of the metadata attribute. Each name MUST exactly
  match one appearing in the "Attributes" sheet. Some names MAY be omitted.
- In Column B, the actual metadata. These may be empty.

TEMPLATE
========

To add information about additional data flows not included in existing sheets
(above), you can copy and rename this sheet.
"""


def _header(ws: "Worksheet", *columns: Tuple[str, int]) -> None:
    """Write header columns and format their style and width."""
    for column, (value, width) in enumerate(columns, start=1):
        cell = cast("Cell", ws.cell(row=1, column=column, value=value))
        cell.style = "header"
        ws.column_dimensions[cell.column_letter].width = width


def add_readme(wb: "Workbook") -> None:
    """Add a "README" sheet to `wb`."""
    ws = wb.create_sheet("README")

    _header(ws, ("Transport Data Commons (TDC) metadata", 72))
    ws["A3"] = README_TEXT


def add_attributes(wb: "Workbook", msd: "v21.MetadataStructureDefinition"):
    """Add an "Attributes" sheet to `wb` listing the metadata attributes from `msd`."""
    ws = wb.create_sheet("Attributes")

    _header(
        ws,
        ("Name", 20),  # "Element name" in Eurostat
        ("Description", 72),  # Not present in Eurostat
        ("ID", 20),  # "Element code" in Eurostat
    )

    for row, attribute in enumerate(msd.report_structure["ALL"].components, start=2):
        concept = attribute.concept_identity
        ws.cell(row=row, column=1, value=concept.name.localized_default()).style = "top"
        ws.cell(row=row, column=2, value=concept.description.localized_default())
        ws.cell(row=row, column=3, value=attribute.id).style = "top"


def add_template(wb: "Workbook", msd: "v21.MetadataStructureDefinition"):
    """Add a "TEMPLATE" sheet to `wb` with a metadata template."""
    ws = wb.create_sheet("TEMPLATE")

    _header(
        ws,
        ("Attribute name", 20),  # "Concept name" in Eurostat
        ("Value", 72),  # "Concept value" in Eurostat
    )

    for row, attribute in enumerate(msd.report_structure["ALL"].components, start=2):
        concept = attribute.concept_identity
        ws.cell(row=row, column=1, value=concept.name.localized_default()).style = "top"
        ws.cell(row=row, column=2, value="---")


def getdefault(is_: "common.ItemScheme", other: "common.Item") -> "common.Item":
    """Return an item from `is_` matching `other`.

    Several methods are attempted to match `other` with an existing item:

    1. ID of `other` is identical to that of an existing item.
    2. Transformed ID of `other`—in upper case, " " replaced with "_" is identical to
       that of an existing item.
    3. ID of `other` is in the annotation ``tdc-aka``

    """
    # Exact match on ID or transformed ID
    for candidate in (other.id, other.id.upper().replace(" ", "_")):
        try:
            return is_[candidate]
        except KeyError:
            pass

    # Iterate over existing items
    for item in is_:
        # Eval the annotation "tdc-aka" for a list of alternate IDs for the item
        if aka := item.eval_annotation(id="tdc-aka"):
            if other.id in aka:
                return item

    # Still no match; create the item
    return is_.setdefault(id=other.id)


def make_workbook(name="sample.xlsx") -> None:
    """Generate a :class:`openpyxl.Workbook` for exchange of metadata."""
    from openpyxl import Workbook
    from openpyxl.styles import Alignment, Font, NamedStyle, PatternFill

    from transport_data.org.metadata import get_msd

    wb = Workbook()

    # Delete the default sheet
    assert wb.active
    wb.remove(wb.active)

    # Create two named styles
    header = NamedStyle(name="header")
    header.fill = PatternFill("solid", fgColor="000000")
    header.font = Font(bold=True, color="ffffff", name="Calibri")
    wb.add_named_style(header)

    top = NamedStyle(name="top")
    top.alignment = Alignment(vertical="top", wrap_text=True)
    top.font = Font(name="Calibri")
    wb.add_named_style(top)

    # Generate the metadata structure definition
    msd = get_msd()

    # Add sheets
    add_readme(wb)
    add_attributes(wb, msd)
    add_template(wb, msd)

    # Save the file
    wb.save(name)


def parse_dimension(value: str) -> List[v21.Concept]:
    """Parse the description of a dimension from `value`.

    Supported values include:

    1. Multiple lines, with each line beginning "- ".
    2. A single line, with dimensions separated by ", ".
    3. A single dimension ID.
    """
    # Partial regular expressions for a dimension
    entry = r"(?P<id>.+?)(?: \((?P<description>[^\)]*)\))?"

    # Split `value` into potentially multiple values; separate dimension IDs from
    # description/annotation
    parts = []
    if matches := re.findall(f"^- {entry}$", value, flags=re.MULTILINE):
        # Multiple lines, with each line beginning "- "
        parts.extend(matches)
    elif matches := re.findall(f"{entry}(?:, |$)", value):
        # Single line, with dimensions separated by ", "
        # TODO Check behaviour if the ", " is within parentheses
        parts.extend(matches)
    elif 0 == len(parts):  # pragma: no cover
        # None of the above → a single dimension label
        parts.append(value)

    # Convert to a list of Concept objects
    return [
        v21.Concept(id=id_.strip(), name=id_, description=description)
        for id_, description in parts
    ]


def read_workbook(
    path: "pathlib.Path",
) -> tuple["v21.MetadataSet", "v21.ConceptScheme"]:
    """Read a metadata set from the workbook at `path`."""
    from openpyxl import load_workbook

    from transport_data.org.metadata import get_cs_common, get_msd

    wb = load_workbook(path)
    # Generate/retrieve the metadata structure definition
    msd = get_msd()

    mds = v21.MetadataSet(structured_by=msd)

    # Create a shared concept scheme for the concepts referenced by dimensions
    # TODO Collect, maybe with get_msd()
    cs_dims = get_cs_common()

    for ws in wb.worksheets:
        # Skip information sheets generated by methods in this file
        if ws.title in ("README", "Attributes", "TEMPLATE"):
            continue

        if r := read_worksheet(ws, msd, cs_dims):
            mds.report.append(r)

    return mds, cs_dims


def read_worksheet(
    ws: "Worksheet",
    msd: "v21.MetadataStructureDefinition",
    cs_dims: "v21.ConceptScheme",
) -> "v21.MetadataReport | None":
    """Read a metadata report from the worksheet `ws`.

    Parameters
    ----------
    msd :
       Metadata structure definition.
    """
    from transport_data import STORE
    from transport_data.org.metadata import _get

    # Mapping from names (not IDs) to MetadataAttributes
    mda_for_name: dict[str, "common.MetadataAttribute"] = {
        str(mda.concept_identity.name): mda
        for mda in msd.report_structure["ALL"].components
    }

    # Create the target of the report: a data flow definition
    # TODO Expand this DFD and its associated data structure definition
    df_id_from_title = ws.title
    ma_args: "MAKeywords" = dict(
        id=ws.title, maintainer=msd.maintainer, version="1.0.0"
    )
    dfd = v21.DataflowDefinition(**ma_args)
    dsd = v21.DataStructureDefinition(**ma_args)
    dsd.measures.getdefault(id="OBS_VALUE")
    dsd.attributes.getdefault(id="COMMENT")
    dfd.structure = dsd

    # Create objects to associate the metadata report with the data flow definition
    iot = v21.IdentifiableObjectTarget()
    tok = v21.TargetObjectKey(
        key_values={"DATAFLOW": v21.TargetIdentifiableObject(value_for=iot, obj=dfd)}
    )

    # Create the report itself
    mdr = v21.MetadataReport()
    mdr.attaches_to = tok

    # Reference to the MetaDataAttribute describing the current row
    mda = common.MetadataAttribute()
    dimension_concepts = []

    # Iterate over rows in the worksheet, skipping the first
    for row in ws.iter_rows(min_row=2):
        try:
            # Column B: value in the row
            ra_value = row[1].value

            if ra_value is None:
                continue
        except IndexError:  # pragma: no cover
            log.warning(
                f"Sheet {df_id_from_title!r} has only < 2 columns in the first row; skip"
            )
            return None

        # Column A: name of the metadata attribute
        mda_name = row[0].value

        # Identify the MDA
        # NB if `mda_name` is none, then `mda` retains the value found on the previous
        #    row. This allows e.g. multiple rows to give values for DIMENSION
        # TODO Protect against other malformed data.
        mda = mda_for_name.get(str(mda_name), mda)

        if mda and mda.id == "DIMENSION":
            # Parse 1 or more dimension(s) and add to the DSD
            dimension_concepts.extend(parse_dimension(str(ra_value)))
        else:
            # Store as OtherNonEnumeratedAttributeValue
            # TODO Use EnumeratedAttributeValue, once code lists are available
            #      corresponding to dimensions
            ra = v21.OtherNonEnumeratedAttributeValue(
                value=str(ra_value).strip(), value_for=mda
            )

            # Attend the reported attribute to the report
            mdr.metadata.append(ra)

    # Basic checks
    df_id_from_cell = _get(mdr, "DATAFLOW")
    if not df_id_from_cell:
        log.warning(f"Sheet {df_id_from_title!r} does not identify a data flow; skip")
        return None

    update_dimension_descriptor(dsd, cs_dims, *dimension_concepts)

    # Store the DSD and DFD
    STORE.set(dsd)
    STORE.set(dfd)

    return mdr


def update_dimension_descriptor(
    dsd: "v21.DataStructureDefinition", cs_dims: "v21.ConceptScheme", *concepts
) -> None:
    """Update the DimensionDescriptor of `dsd` with `concepts`."""
    for dc in concepts:
        # Identify the concept in `cs_dims` with the same ID
        c = getdefault(cs_dims, dc)

        # Construct annotations
        anno = [v21.Annotation(id="tdc-description", text=dc.description)]
        if c.id != dc.id:
            anno.append(v21.Annotation(id="tdc-original-id", text=dc.id))

        dsd.dimensions.getdefault(id=c.id, concept_identity=c, annotations=anno)



"""Utilities for interacting with CKAN instances via their Action API.

The :mod:`ckanapi` package (`PyPI <https://pypi.org/project/ckanapi/>`_,
`GitHub <https://github.com/ckan/ckanapi>`_), maintained by the CKAN organization,
provides Pythonic access to the “CKAN Action API” (
`documentation <https://docs.ckan.org/en/latest/api/index.html>`_), which itself exposes
nearly all of the functionality available through the CKAN web interface and the TDC
front-end.

The :class:`~.ckan.Client` class is a wrapper around the :class:`ckanapi.RemoteCKAN`
class that provides conveniences used by other code in :mod:`transport_data`.
"""

from functools import partialmethod
from importlib.metadata import version
from itertools import count
from typing import TYPE_CHECKING, ClassVar, TypeVar
from warnings import filterwarnings

if TYPE_CHECKING:
    import pathlib

    from ckanapi import RemoteCKAN

# Work around https://github.com/ckan/ckanapi/pull/218 in ckanapi <= 4.8
filterwarnings(
    "ignore", "pkg_resources is deprecated", DeprecationWarning, "ckanapi.version"
)
filterwarnings(
    "ignore", ".*pkg_resources.declare_namespace", DeprecationWarning, "pkg_resources"
)

T = TypeVar("T", bound="ModelProxy")


class ModelProxy:
    """Simple proxy for a CKAN object/model.

    :mod:`ckan` itself is a Python package, but is a ‘heavy’ dependency—a large package
    with many dependencies. ModelProxy allows to interact with the different classes of
    CKAN objects based on the JSON data returned by the CKAN Action API, without a
    dependency on :mod:`ckan` itself.
    """

    #: Object 'name'. In the CKAN data model, this is often a snake-case string like
    #: "name-of-data-set".
    name: str | None = None

    #: Object 'id'. In the CKAN data model, this is often a :mod:`uuid` of some sort.
    id: str | None = None

    _collections: ClassVar[dict[str, tuple[type, str]]] = dict()

    def __init__(self, data: dict | None = None, **kwargs) -> None:
        self.__dict__.update(data or {})
        self.__dict__.update(kwargs)
        self._process_collections()

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.__dict__ == other.__dict__

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        return (
            f"<CKAN {type(self).__name__} "
            + (repr(self.name) if self.name else "(no name)")
            + f" | {len(self.__dict__) - 1} fields | id="
            + self.__dict__.get("id", "(none)")
            + ">"
        )

    @classmethod
    def from_file(cls: type[T], path: "pathlib.Path") -> T:
        """Construct a new instance from a file `path`."""
        import json

        with open(path) as f:
            return cls(json.load(f))

    def asdict(self) -> dict:
        """Return the original dictionary of object data."""
        return self.__dict__.copy()

    def get(self, name: str):
        """Return the value of a single field."""
        return self.__dict__.get(name)

    def get_item(self, name: str, index: int | None = None):
        """Get a member of a collection."""
        data = self.__dict__[name][index]
        cls = get_class(name)
        assert cls
        return cls(data)

    def _process_collections(self) -> None:
        """Convert the :attr:`_collections` to the designated types."""
        for name, (t_collection, cls_name) in self._collections.items():
            cls = get_class(cls_name)
            assert cls is not None

            data = self.__dict__.get(name)
            if data is None:
                continue

            self.__dict__[name] = [(cls(x) if isinstance(x, dict) else x) for x in data]

    def update(self, data: dict) -> None:
        """Update part or all of the object data."""
        # Check some items
        if self.name and data.get("name", self.name) != self.name:
            raise ValueError(f"Cannot update with {data['name']!r} != {self.name=!r}")
        elif self.id and data.get("id", self.id) != self.id:
            raise ValueError(f"Cannot update with {data['id']!r} != {self.id=!r}")
        self.__dict__.update(data)
        self._process_collections()


def get_class(name: str) -> type[ModelProxy] | None:
    """Return a :class:`.ModelProxy` subclass given `name`."""
    glb = globals()
    for candidate in (
        name.title(),
        name.rstrip("s").title(),
        {"member_role": "MemberRole"}.get(name.rstrip("s"), ""),
    ):
        try:
            return glb[candidate]
        except KeyError:
            pass
    return None


class Group(ModelProxy):
    """Proxy for `ckan.model.Group
    <https://github.com/ckan/ckan/blob/master/ckan/model/group.py>`_.
    """


class Organization(Group):
    """'Organization' is a synonym for 'Group'."""

    # NB this is a subclass instead of `Organization = Group` so that type(…).__name__
    #    gives 'organization'

    _collections = {
        "packages": (list, "Package"),
    }

    def __repr__(self) -> str:
        lines = [super().__repr__()]
        lines.extend(f"  {p!r}" for p in (self.get("packages") or []))
        return "\n".join(lines)


class MemberRole(ModelProxy):
    """Proxy for the CKAN 'MemberRole' model.

    .. todo:: Add a link to the proxied class.
    """


class License(ModelProxy):
    """Proxy for `ckan.model.License
    <https://github.com/ckan/ckan/blob/master/ckan/model/license.py>`_.
    """


class Package(ModelProxy):
    """Proxy for `ckan.model.Package
    <https://github.com/ckan/ckan/blob/master/ckan/model/package.py>`_.
    """


class Resource(ModelProxy):
    """Proxy for `ckan.model.Resource
    <https://github.com/ckan/ckan/blob/master/ckan/model/resource.py>`_.
    """


class Tag(ModelProxy):
    """Proxy for the CKAN 'Tag' model.

    The source code for the proxied class is `here
    <>`_.
    """


class User(ModelProxy):
    """Proxy for `ckan.model.User
    <https://github.com/ckan/ckan/blob/master/ckan/model/user.py>`_.
    """


class Client:
    """Wrapper around :class:`ckanapi.RemoteCKAN`.

    This provides features such as:

    - Iterating over calls with rate limits.
    - Caching results, including combined results from multiple calls.
    - Converting return values to instances of :class:`ModelProxy` subclasses.
    """

    _api: "RemoteCKAN"
    _cache: dict
    id: str

    def __init__(self, address: str, id: str) -> None:
        from ckanapi import RemoteCKAN

        self.id = id

        # Construct a user-agent string
        kw: dict[str, str | None] = dict(
            user_agent=f"transport_data/{version('transport_data')} "
            "(+https://docs.transport-data.org)"
        )

        # Fetch the user's API token, if any
        try:
            import keyring

            kw.update(
                apikey=keyring.get_password("transport-data", f"api-token-{self.id}")
            )
        except ImportError:
            pass

        self._api = RemoteCKAN(address, **kw)
        self._cache = dict(package=dict())

    def __getattr__(self, name: str):
        return getattr(self._api.action, name)

    def list_action(
        self,
        kind: str,
        limit: int | None = None,
        max: int | None = None,
        **kwargs,
    ) -> list:
        """Call the ``{kind}_list`` API endpoint.

        Parameters
        ----------
        kind
            String identifying the kind of CKAN object to fetch.
        limit
            Number of results to fetch in a single query.
        max
            Maximum number of results to fetch.
        """
        cls = get_class(kind)
        assert cls

        limit = limit or 100

        c = self._cache[f"{kind}_list"] = list()

        data_dict = {"limit": limit}
        data_dict.update(kwargs)

        for i in count():
            data_dict.update(offset=i * limit)
            result = self._api.call_action(f"{kind}_list", data_dict=data_dict)
            c.extend(result)
            if len(result) < limit or (max and max < (i + 1) * limit):
                break
        return [(cls(name=v) if isinstance(v, str) else cls(v)) for v in c]

    # Use list_action() to invoke certain CKAN API endpoints
    group_list = partialmethod(list_action, "group")
    license_list = partialmethod(list_action, "license")
    member_roles_list = partialmethod(list_action, "member_roles")
    organization_list = partialmethod(list_action, "organization")
    package_list = partialmethod(list_action, kind="package")
    tag_list = partialmethod(list_action, kind="tag")
    user_list = partialmethod(list_action, kind="user")

    def show_action(self, obj_or_id: str | dict | T, _cls: type[T], **kwargs) -> T:
        """Call the ``{kind}_show`` API endpoint.

        If `obj_or_id` is an instance of :class:`.ModelProxy`, its
        :attr:`~ModelProxy.name` or :attr:`~.ModelProxy.id` is used for the query, and
        the same instance is updated with the response data and returned. If `obj_or_id`
        is :class:`str` or :class:`dict` a new instance of a :class:`.ModelProxy`
        subclass is returned.

        The return value is cached.
        """
        if isinstance(obj_or_id, dict):
            kw: dict[str, str | None] = obj_or_id
            kw.update(id=kw.pop("name", kw.get("id", "")))
            result: T | None = _cls()
        else:
            kw = dict()

        if isinstance(obj_or_id, str) or obj_or_id is None:
            # Query using the "id" keyword
            kw.update(id=obj_or_id)
            # Create a new result object according to the type of
            result = None
        elif isinstance(obj_or_id, _cls):
            if obj_or_id.id:
                kw.update(id=obj_or_id.id)
            elif obj_or_id.name:
                kw.update(id=obj_or_id.name)
            result = obj_or_id

        kw.update(kwargs)

        kind = _cls.__name__.lower()

        response = self._api.call_action(f"{kind}_show", kw)

        # TODO Handle whatever exception is raised for an invalid `package`
        if result is None:
            result = _cls(response)
        else:
            result.update(response)

        # Update cache by both UUID and name
        self._cache[result.id] = result
        self._cache.setdefault(kind, dict())
        self._cache[kind][result.name] = result

        return result

    # Use show_action() to invoke certain CKAN API endpoints
    # TODO Extend to cover all "*_show" endpoints.
    organization_show = partialmethod(show_action, _cls=Organization)
    package_show = partialmethod(show_action, _cls=Package)
    tag_show = partialmethod(show_action, _cls=Tag)
